/*
FILE:     CifSchemaMap.h
*/
/*
VERSION:  4.017
*/
/*
DATE:     2/8/2008
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2008 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               RCSB PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/


/*!
** \file CifSchemaMap.h
**
** \brief Header file for DbLoader related classes.
*/


#ifndef CIFSCHEMAMAP_H
#define CIFSCHEMAMAP_H


#include <string>

#include "CifFile.h"


typedef struct _attribInfo
{
    int iIndex;
    int iNull;
    int iWidth;
    int iPrecision;
    unsigned int iMaxWidth; 
    string dType;
    int iTypeCode;
} ATTRIBINFO;


typedef struct
{
    string attribName;
    string dataType;
    string indexFlag;
    string nullFlag;
    string width;
    string precision;
    string populated;   // Not taken from the file, but constructed afterwards
} AttrInfo;


/**
**  \class SchemaMapping
**
**  \brief Public class that respresents schema mapping configuration.
**
**  This class encapsulates schema mapping functionality. Schema mapping
**  defines how mmCIF data gets converted into the database data. Schema
**  mapping information is stored in a number of configuration tables
**  that are located in a configuration CIF file. This class provides methods
**  for reading the schema mapping information from the ASCII CIF file or
**  serialized (binary) configuration CIF file, for revising the schema
**  mapping information (based on the data), updating the schema mapping
**  information, accessing the attributes information.
*/
class SchemaMapping
{

  public:
    // TODO - All these constants can be made
    // private, except _TYPE_CODE_TEXT that is used outside of this
    // class. Try to fix that and move all these constants to private
    static const int _TYPE_CODE_INT        = 1;
    static const int _TYPE_CODE_FLOAT      = 2;
    static const int _TYPE_CODE_STRING     = 3;
    static const int _TYPE_CODE_TEXT       = 4;
    static const int _TYPE_CODE_DATETIME   = 5;
    static const int _TYPE_CODE_BIGINT     = 6;

    static const int _MAX_LINE_LENGTH = 255;

    /**
    **  Constructs a schema mapping object.
    **  
    **  \param[in] schemaFile - optional parameter that indicates the name of
    **    the ASCII CIF file, which holds the schema mapping configuration
    **    information. If specified, \e schemaFile file is parsed and the
    **    schema mapping configuration information is taken from it.
    **  \param[in] schemaFileOdb - optional parameter that indicates the name
    **    of the serialized CIF file, which holds the schema mapping
    **    configuration information. If \e schemaFile is not specified,
    **    \e schemaFileOdb is de-serialized and the schema mapping
    **    configuration information is taken from it. If both \e schemaFile
    **    and \e schemaFileOdb are specified, \e schemaFile is parsed,
    **    the schema mapping configuration information is taken from it and
    **    then serialized to \e schemaFileOdb file, at the end of processing.
    **  \param[in] verbose - optional parameter that indicates whether
    **    logging should be turned on (if true) or off (if false).
    **    If \e verbose is not specified, logging is turned off.
    **
    **  \return Not applicable
    **
    **  \pre \e schemaFile and \e schemaFileOdb cannot both be empty.
    **
    **  \post None
    **
    **  \exception: None
    */
    SchemaMapping(const string& schemaFile = String::Empty,
      const string& schemaFileOdb = String::Empty, bool verbose = false);

    /**
    **  Destructs a schema mapping object, by releasing all the used resources.
    **
    **  \param: Not applicable
    **
    **  \return Not applicable
    **
    **  \pre None
    **
    **  \post None
    **
    **  \exception: None
    */
    virtual ~SchemaMapping();

    /**
    **  Sets the schema revising mode. Schema revising is a process where
    **  the existing schema mapping information is being updated, by examining
    **  the attributes of the converted mmCIF file. Schema revising is needed
    **  in order to increase the length of database attributes in database
    **  schema/data file so that the data can be fully stored in the database.
    **
    **  \param[in] mode - optional parameter that indicates whether
    **    the schema revising should be turned on (if true) or off (if false).
    **    If \e mode is not specified, schema revising is turned on.
    **
    **  \return None
    **
    **  \pre None
    **
    **  \post None
    **
    **  \exception: None
    */
    void SetReviseSchemaMode(bool mode = true);

    /**
    **  Retrieves schema revising mode.
    **
    **  \param: None
    **
    **  \return true - if schema revising is turned on
    **  \return false - if schema revising is turned off
    **
    **  \pre None
    **
    **  \post None
    **
    **  \exception: None
    */
    bool GetReviseSchemaMode();

    /**
    **  Writes the revised schema mapping information to a file.
    **  
    **  \param[in] revisedSchemaFile - indicates the name of
    **    the ASCII CIF file, to which the revised schema mapping information
    **    will be written to.
    **
    **  \return None
    **
    **  \pre None
    **
    **  \post None
    **
    **  \exception: None
    */
    void ReviseSchemaMap(const string& revisedSchemaFile);


    /**
    **  Updates the existing schema mapping object, with the revised
    **  information from the specified schema mapping object.
    **
    **  \param[in] revSchMap - reference to a revised schema mapping object.
    **
    **  \return None
    **
    **  \pre None
    **
    **  \post None
    **
    **  \exception: None
    */
    void updateSchemaMapDetails(SchemaMapping& revSchMap);

    /**
    **  In the specified CIF file object, creates tables that are specified in
    **  the schema mapping object.
    **
    **  \param[in] cifFile - reference to a CIF file object that will hold
    **    the data.
    **  \param[in] blockName - the name of the data block
    **
    **  \return None
    **
    **  \pre None
    **
    **  \post None
    **
    **  \exception: None
    */
    void CreateTables(CifFile& cifFile, const string& blockName);

    /**
    **  Retrieves names of all the tables in each data block, defined in
    **  schema mapping object.
    **
    **  \param[out] tableNames - retrieved table names
    **
    **  \return None
    **
    **  \pre None
    **
    **  \post None
    **
    **  \exception: None
    */
    void GetTableNames(vector<string>& tableNames);

    ISTable* CreateTableInfo();
    ISTable* CreateColumnInfo();

    void GetAttributeNames(vector<string>& attributes,
      const string& tableName);
    void GetAttributesInfo(vector<AttrInfo>& attrInfo,
      const string& tableName);
    void GetMappedAttributesInfo(vector<vector<string> >& mappedAttrInfo,
      const string& tableName);

    void GetMappedConditions(vector<vector<string> >& mappedConditions,
      const string& tableName);

    void GetTableNameAbbrev(string &abbrev, string target);
    void GetAttributeNameAbbrev(string &abbrev, string targetTable,
      string targetAttribute);

    void getSchemaMapDetails(const string& tableName,
      const string& attribName, string& dataType,
      string& indexFlag, string& nullFlag,
      string& width, string& precision, string& populated);

    void UpdateAttributeDef(const string& tableName, const string& columnName,
      int type, int iWidth, int newWidth);

    void GetAttributeInfo(vector<ATTRIBINFO>& aI, ISTable *tIn);
    void GetAllAttributeInfo(vector<ATTRIBINFO>& aI, string tableName);

    static bool AreAttributesMandatory(const vector<string>& row,
      const vector<ATTRIBINFO>& aI);
    static bool IsTablePopulated(const vector<AttrInfo>& attrInfo);
    void GetMasterIndexAttribName(string& masterIndexAttribName);
    static int CheckIndexAttributes(const vector<string>& row,
      const vector<ATTRIBINFO>& aI);

  private:
    bool _verbose;

    string _schemaFile;    // Schema definition file name (mmCIF)
    string _schemaFileOdb; // Schema definition file name (odb)

    CifFile* _fobjS;

    ISTable* _tableSchema;
    ISTable* _attribSchema;
    ISTable* _schemaMap;
    ISTable* _mapConditions;
    ISTable* _tableAbbrev;    
    ISTable* _attributeAbbrev;

    bool _reviseSchemaMode;
    int  _compatibilityMode;

    void _AssignAttribIndices(void);
    void GetGroupNames(vector<string>& groupNames);

    void Clear();
};


/**
**  \class Db
**
**  \brief Base class for all databases.
**
**  This class encapsulates general characteristics of each database. It
**  contains methods for generic database related operations and virtual
**  methods, that are to be specialized in concrete databases.
*/
class Db
{
  public:
    static const string DB_DEFAULT_NAME;

    SchemaMapping& _schemaMapping;

    Db(SchemaMapping& schemaMapping, const string& dbName = DB_DEFAULT_NAME);
    virtual ~Db();

    void SetUseOnlyPopulated(bool mode=true);
    bool GetUseOnlyPopulated();

    void SetAppendFlag(const bool appendFlag);
    bool GetAppendFlag();

    void SetFieldSeparator(const string& fieldSeparator);
    void SetRowSeparator(const string& rowSeparator);

    string GetCommandTerm();

    string GetFieldSeparator();
    string GetRowSeparator();

    virtual void GetStart(string& start);

    virtual void WriteSchemaStart(ostream& io);
    void WriteDeleteTable(ostream& io, const string& table,
      const string& where, const string& what);

    virtual void DropTableSql(ostream& io, const string& tableNameDb);

    virtual const string& GetExec();
    virtual const string& GetExecOption();
    virtual const string& GetUserOption();
    virtual const string& GetPassOption();

    virtual const string& GetEnvDbUser();
    virtual const string& GetEnvDbPass();

    virtual const string& GetConnect();
    virtual const string& GetTerminate();
    virtual const string& GetDbCommand();

    virtual void WriteLoad(ostream& io);

    const string& GetDataLoadingFileName();

    virtual void WriteLoadingStart(ostream& io);
    virtual void WriteLoadingEnd(ostream& io);
    virtual void WriteLoadingTable(ostream& io, const string& tableName,
      const string& path);

    virtual void WritePrint(ostream& io, const string& tableNameDb);

    virtual void GetChar(string& dType, const unsigned int width);
    virtual void GetFloat(string& dType);
    virtual void GetText(string& dType, const unsigned int width);
    virtual void GetDate(string& dType);
    virtual void WriteNull(ostream& io, const int iNull,
      const unsigned int curr, const unsigned int attSize);
    virtual void WriteTableIndex(ostream& io, const string& tableNameDb,
      const vector<string>& indexList);

    const string& GetBcpStringDelimiter();
    virtual void WriteBcpDoubleQuotes(ostream& io);

    virtual void WriteNewLine(ostream& io, bool special = false);

    bool IsFirstTextNewLineSpecial();

    virtual void ConvertDate(string& dbDate, const string& cifDate);
    virtual void ConvertTimestamp(string& dbTimestamp,
      const string& cifTimestamp);

  protected:
    bool _useOnlyPopulated;

    bool _appendFlag;

    // Field and row separators for compact output (eg. BCP)
    string _fieldSeparator; 
    string _rowSeparator;   

    string _cmdTerm;            // SQL command terminator.

    // Used in assigning permissions when exporting schema in SQL. 
    string _dbName; // Target database name 

    string _exec;
    string _execOption;
    string _userOption;
    string _passOption;

    string _connect;
    string _terminate;
    string _dbCommand;
    string _envDbUser;
    string _envDbPass;

    string _dataLoadingFileName;

    string _bcpStringDelimiter;

    bool _firstTextNewLineSpecial;

  private:
    static const string _SCRIPT_LOADING_FILE;

};


/**
**  \class DbOracle
**
**  \brief Oracle database class.
**
**  This class represents an Oracle database.
*/
class DbOracle : public Db
{

  // Oracle 8.1.6 SQLLDR

  public:
    DbOracle(SchemaMapping& schemaMapping,
      const string& dbName = DB_DEFAULT_NAME);
    ~DbOracle();

    void WriteSchemaStart(ostream& io);

    void WriteLoadingStart(ostream& io);
    void WriteLoadingTable(ostream& io, const string& tableName,
      const string& path);

    void GetDate(string& dType);
    void GetText(string& dType, const unsigned int width);

    void WriteNull(ostream& io, const int iNull,
      const unsigned int curr, const unsigned int attSize);
    void WriteTableIndex(ostream& io, const string& tableNameDb,
      const vector<string>& indexList);

    void WriteNewLine(ostream& io, bool special = false);
};


/**
**  \class DbDb2
**
**  \brief DB2 database class.
**
**  This class represents a DB2 database.
*/
class DbDb2 : public Db
{

  public:
    DbDb2(SchemaMapping& schemaMapping,
      const string& dbName = DB_DEFAULT_NAME);
    ~DbDb2();


    void GetStart(string& start);

    void WriteSchemaStart(ostream& io);

    void WriteLoadingStart(ostream& io);
    void WriteLoadingEnd(ostream& io);
    void WriteLoadingTable(ostream& io, const string& tableName,
      const string& path);

    void GetFloat(string& dType);
    void GetDate(string& dType);
    void GetText(string& dType, const unsigned int width);
    void WriteNull(ostream& io, const int iNull,
      const unsigned int curr, const unsigned int attSize);

    void WriteTableIndex(ostream& io, const string& tableNameDb,
      const vector<string>& indexList);

    void WriteBcpDoubleQuotes(ostream& io);
};


/**
**  \class DbMySql
**
**  \brief MySQL database class.
**
**  This class represents a MySQL database.
*/
class DbMySql : public Db
{

  public:
    DbMySql(SchemaMapping& schemaMapping,
      const string& dbName = DB_DEFAULT_NAME);
    ~DbMySql();

    void GetStart(string& start);

    void DropTableSql(ostream& io, const string& tableNameDb);

    void WriteLoad(ostream& io);
    void WriteLoadingStart(ostream& io);
    void WriteLoadingTable(ostream& io, const string& tableName,
      const string& path);

    void WriteTableIndex(ostream& io, const string& tableNameDb,
      const vector<string>& indexList);

    void WriteNull(ostream& io, const int iNull,
      const unsigned int curr, const unsigned int attSize);

    void WriteNewLine(ostream& io, bool special = false);

  private:
    static const string _SQL_LOADING_FILE;

};


/**
**  \class DbSybase
**
**  \brief Sybase database class.
**
**  This class represents a Sybase database.
*/
class DbSybase : public Db
{
  public:
    DbSybase(SchemaMapping& schemaMapping,
      const string& dbName = DB_DEFAULT_NAME);
    ~DbSybase();

    void GetStart(string& start);

    void WriteLoadingStart(ostream& io);
    void WriteLoadingTable(ostream& io, const string& tableName,
      const string& path);

    void WritePrint(ostream& io, const string& tableNameDb);
    void WriteNull(ostream& io, const int iNull,
      const unsigned int curr, const unsigned int attSize);

    void WriteTableIndex(ostream& io, const string& tableNameDb,
      const vector<string>& indexList);

#ifdef VLAD_DATE_OBSOLETE
    void ConvertDate(string& dbDate, const string& cifDate);
#endif
};


/**
**  \class DbOutput
**
**  \brief Base class for all output formats.
**
**  This class encapsulates general characteristics of each kind of output.
**  It contains virtual methods for generating the schema, data and loading
**  scripts.
*/
class DbOutput
{

  public:
    Db& _db;

    DbOutput(Db& db);
    virtual ~DbOutput();

    virtual void WriteSchema(const string& path = String::Empty);
    virtual void WriteDataLoadingScripts(const string& path = String::Empty);
    virtual void WriteData(Block& block, const string& path = String::Empty);

    void SetInputFile(const string& inpFile);

    const string& GetCommandScriptName();

  protected:
    static const string _DATA_LOADING_SCRIPT;

    string _SCHEMA_FILE;

    string _INPUT_FILE;

    string _stringDelimiter;
    vector<char> _specialChars;

    string _dateDelimiter;
    vector<char> _specialDateChars;

    string _itemSeparator; 
    string _rowSeparator; 

    void WriteDbExec(ostream& io, const string& fileName,
      const unsigned int indentLevel = 0);
    void WriteDbExecOnly(ostream& io, const string& fileName,
      const unsigned int indentLevel = 1);

    void WriteHeader(ostream& io);

    void _FormatNumericData(ostream &io, const string &cs);
    void _FormatStringData(ostream &io, const string &cs,
      unsigned int maxWidth);
    void _FormatTextData(ostream &io, const string &cs);
    void _FormatDateData(ostream &io, const string &cs, unsigned int maxWidth);

    void _FormatData(ostream &io, const string &cs,
      ATTRIBINFO& attribInfo, const bool noRevise);

    bool IsSpecialChar(const char& character);
    bool IsSpecialDateChar(const char& character);

    virtual void WriteEmptyNumeric(ostream& io);
    virtual void WriteEmptyString(ostream& io);
    virtual void WriteSpecialChar(ostream& io, const char& specChar);

    virtual void WriteEmptyDate(ostream& io);
    virtual void WriteSpecialDateChar(ostream& io, const char& specDateChar);

    virtual bool IsFirstTextNewLineSpecial();
    virtual void WriteNewLine(ostream& io, bool special = false);

    virtual void GetTableStart(string& tableStart, const string& tableName);
    virtual void GetTableEnd(string& tableEnd);
    const string& GetItemSeparator();
    const string& GetRowSeparator();

    virtual void _WriteTable(ostream& io, ISTable* tIn, bool noRevise = false);
    void GetMasterIndexAttribValue(string& masterIndexAttribValue,
      Block& block, const string& masterIndexAttribName,
      const vector<string>& tableNames);

  private:
    static void _FormatStringDataSql(ostream &io, const string &cs,
      unsigned int maxWidth);

};


/**
**  \class BcpOutput
**
**  \brief BCP output class.
**
**  This class represents a BCP output.  It re-implements methods
**  for data and loading scripts generation.
*/
class BcpOutput : public DbOutput
{
  public:
    BcpOutput(Db& db);
    virtual ~BcpOutput();

    void WriteDataLoadingScripts(const string& path = String::Empty);
    void WriteData(Block& block, const string& path = String::Empty);

  private:
    static const string _DATA_DELETE_FILE;

    void WriteDataLoadingScript(const string& path);
    void WriteDataLoadingFile(const string& path = String::Empty);

    void WriteDelete(ostream& io);

    void WriteEmptyString(ostream& io);

    void WriteSpecialDateChar(ostream& io, const char& specDateChar);
};
 

/**
**  \class SqlOutput
**
**  \brief SQL output class.
**
**  This class represents an SQL output. It re-implements methods
**  for schema, data and loading scripts generation.
*/
class SqlOutput : public DbOutput
{

  public:
    SqlOutput(Db& db);
    virtual ~SqlOutput();

    void WriteSchema(const string& path = String::Empty);
    void WriteDataLoadingScripts(const string& path = String::Empty);
    void WriteData(Block& block, const string& path = String::Empty);

  protected:
    void WriteEmptyNumeric(ostream& io);
    bool IsFirstTextNewLineSpecial();
    void WriteNewLine(ostream& io, bool special = false);
    void GetTableStart(string& tableStart, const string& tableName);
    void GetTableEnd(string& tableEnd);

  private:
    static const unsigned int _MAX_SQL_NAME_LENGTH = 60;
    static const string _SCHEMA_LOADING_SCRIPT;
    static const string _SCHEMA_DELETE_FILE;
    static const string _DATA_FILE;

    void WriteSqlScriptSchemaInfo(ostream& io);
    void WriteDataLoadingScript(const string& path);

    void CreateTableSql(ostream& io, const string& tableName,
      const vector<AttrInfo>& attrInfo);

    void WriteAuxTables(ostream& io, ISTable* infoP,
      const vector<string>& tableNames);

    void WriteEmptyString(ostream& io);
    void WriteEmptyDate(ostream& io);
};


/**
**  \class XmlOutput
**
**  \brief XML output class.
**
**  This class represents an XML output. It re-implements methods
**  for schema and data writing.
*/
class XmlOutput : public DbOutput
{
  public:
    XmlOutput(Db& db);
    virtual ~XmlOutput();

    void WriteSchema(const string& path = String::Empty);
    void WriteData(Block& block, const string& path = String::Empty);

  private:
    static const int _MAX_NAME_LENGTH = 80;
    string _NS_DEFAULT;

    void _CreateComplexTypeXML(ostream &io, const string& tableName);
    void _SetNamespaceDefaultXML(const string& name); 
    void _GetDefaultNamespaceXML(ostream &io); 
    void _WriteTable(ostream& io, ISTable* tIn, bool noRevise = false);

    void _GetAttributeNameAbbrevXML(string &abbrev,
      const string& targetAttribute);
    void _QualifyNameXML(ostream &io,  const string& name); 

    void _GetTableNameAbbrevXML(string &abbrev, string target);

    void WriteSpecialChar(ostream& io, const char& specChar);
};


/**
**  \class DbLoader
**
**  \brief The controller of CIF to DB data conversion process.
**
**  This class controls the conversion process of CIF data to DB loadable data.
**  It provides methods for its construction/destruction, conversion of
**  ASCII CIF files, conversion of serialized (binary) CIF files, conversion of
**  in-memory CIF file objects, setting the working directory. Data conversion
**  methods can be instructed to generate: only the data loading files, only
**  the data loading shell scripts or both the data and scripts. DB server
**  data loading is accomplished by executing the generated loading shell
**  scripts, which utilize the data loading files, obtained during the
**  conversion process.
*/
class DbLoader
{
  public:
    enum eConvOpt
    {
        // Generate only data loading files
        eDATA_ONLY = 0,

        // Generate both data loading files and loading shell scripts
        eDATA_WITH_SCRIPTS,

        // Generate only loading shell scripts
        eSCRIPTS_ONLY
    };

    /**
    **  Constructs a CIF to DB controller object.
    **  
    **  \param[in] schemaMapping - reference to the schema mapping object
    **  \param[in] dbOutput - reference to the database output object
    **  \param[in] verbose - optional parameter that indicates whether
    **    logging should be turned on (if true) or off (if false).
    **    If \e verbose is not specified, logging is turned off.
    **  \param[in] workDir - optional parameter that indicates the working
    **    directory in which the files will be placed. If \e workDir is
    **    not specified, current process working directory is used.
    **
    **  \return Not applicable
    **
    **  \pre None
    **
    **  \post None
    **
    **  \exception: None
    */
    DbLoader(SchemaMapping& schemaMapping, DbOutput& dbOutput,
      bool verbose = false, const string& workDir = String::Empty);

    /**
    **  Destructs a CIF to DB controller object.
    **
    **  \param: Not applicable
    **
    **  \return Not applicable
    **
    **  \pre None
    **
    **  \post None
    **
    **  \exception: None
    */
    virtual ~DbLoader();

    /**
    **  Sets the working directory in which the generated files will be placed.
    **
    **  \param[in] workDir - indicates the working directory in which the
    **    files will be placed.
    **
    **  \return None
    **
    **  \pre None
    **
    **  \post None
    **
    **  \exception: None
    */
    void SetWorkDir(const string& workDir);

    /**
    **  Converts an ASCII CIF file to DB loadable data with/without generating
    **  loading scripts. The specified file is first parsed and then converted.
    **
    **  \param[in] asciiFile - indicates the name of the ASCII CIF file,
    **    which is to be converted.
    **  \param[in] convOpt - indicates data conversion options: generate
    **    data only, generate data with loading scripts, generate loading
    **    scripts only.
    **
    **  \return None
    **
    **  \pre None
    **
    **  \post None
    **
    **  \exception: None
    */
    void AsciiFileToDb(const string& asciiFile, const eConvOpt convOpt);

    /**
    **  Converts a serialized (binary) CIF file to DB loadable data
    **  with/without generating loading scripts. The specified file is
    **  de-serialized (without the need to parse it) and converted.
    **  
    **  \param[in] serFile - indicates the name of the serialized (binary)
    **    CIF file, which is to be converted.
    **  \param[in] convOpt - indicates data conversion options: generate
    **    data only, generate data with loading scripts, generate loading
    **    scripts only.
    **
    **  \return None
    **
    **  \pre None
    **
    **  \post None
    **
    **  \exception: None
    */
    void SerFileToDb(const string& serFile, const eConvOpt convOpt);

    /**
    **  Converts an in-memory CIF file object to DB loadable data
    **  with/without generating loading scripts.
    **  
    **  \param[in] cifFile - a reference to the in-memory CIF file object,
    **    which is to be converted.
    **  \param[in] convOpt - indicates data conversion options: generate
    **    data only, generate data with loading scripts, generate loading
    **    scripts only.
    **
    **  \return None
    **
    **  \pre None
    **
    **  \post None
    **
    **  \exception: None
    */
    void FileObjToDb(CifFile& cifFile, const eConvOpt convOpt);

    /**
    **  Utility method, not part of users public API.
    */
    static unsigned int GetTableColumnIndex(const ISTable& isTable,
      const string& colName);

#ifdef DB_HASH_ID
    void SetHashMode(int mode);
#endif

  private:
    static const string _LOG_FILE;

    string _workDir; // Working directory for all generated files.
    string _INPUT_FILE;

#ifdef DB_HASH_ID
    string _HASH_ID;
    int _hashMode;
#endif

    // Block name of loadable data in mmCIF format
    string _blockName;

    bool _verbose;

    ofstream _log;

    SchemaMapping& _schemaMapping;
    DbOutput& _dbOutput;

    void _LoadBlock(Block& rBlock, Block& wBlock);
    bool _Search(vector<vector<string> >& dMap, const unsigned int iAttr,
      ISTable* isTableP, const string& blockName,
      const vector<string>& cNameMap, const string& sItem,
      const string& sCnd, const string& sFnct);
 
    void _DoFunc(vector<string>& s, const vector<string>& r,
      const string& sFnct);

    void _OpenLog(const string& logName);

    int _GetMapColumnIndex(const vector<string>& cNameMap, const string& vOf);
    void _GetMapColumnValue(string& p, vector<vector<string> >& dMap,
      int iCol, int irow);
    int _GetMapColumnLength(vector<vector<string> >& dMap,   int iCol);

#ifdef VLAD_DATE_OBSOLETE
    void _dformat_1(const char *date, char *odate);
    void _dformat_2(const char *date, char *odate);
    void _dformat_3(const char *date, char *odate, int shortFlag);
    void _dformat_4(const char *date, char *odate);
    void _dformat_5(const char *date, char *odate);
#endif

    void _ReorderName(string& res, char *string, int mode);
    void _ToUpperString(string& aString);
    void _StripString(string& aString, int mode);

    int _CheckNullRow(const vector<string>& row, const vector<ATTRIBINFO>& aI);

#ifdef DB_HASH_ID
    long long pdbIdHash(const string& id);
#endif

    static void CleanString(string& aString);

    void Clear();
};


#endif
