/*
FILE:     CifSchemaMap.C
*/
/*
VERSION:  4.017
*/
/*
DATE:     2/8/2008
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2008 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               RCSB PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/

 
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "CifParserBase.h"
#include "CifSchemaMap.h"

// using std::string::size_type;

// For controlling schema only related operations
const string SqlOutput::_SCHEMA_LOADING_SCRIPT =
  "DB_LOADER_SCHEMA_COMMANDS.csh";
const string SqlOutput::_SCHEMA_DELETE_FILE = "DB_LOADER_SCHEMA_DROP.sql";

// For controlling data loading only related operations
const string DbOutput::_DATA_LOADING_SCRIPT = "DB_LOADER_COMMANDS.csh";

// For deleting existing tables
const string BcpOutput::_DATA_DELETE_FILE = "DB_LOADER_DELETE.sql";

// For data loading
// For BCP loading done via SQL statements
const string DbMySql::_SQL_LOADING_FILE = "DB_LOADER_LOAD.sql";

// For BCP loading done via shell invocations of DB loader
const string Db::_SCRIPT_LOADING_FILE = "DB_LOADER_LOAD_COMMANDS.csh";

// For SQL loading via SQL statements
const string SqlOutput::_DATA_FILE = "DB_LOADER.sql";

const string DbLoader::_LOG_FILE = "SchemaMap.log";

const string Db::DB_DEFAULT_NAME = "msd1";

static void escapeString(string& outStr, const string& inStr);
static void _ConvertDataTypeXML(ostream &io,  string dType);
static void _set_attribute_value_where(ISTable *t,  const string& category,
  const string& attribute, const string& attributeChange,
  const string& valueChange);


const long _MAXFILESIZE = std::numeric_limits<std::streamsize>::max();

static void name_conversion_first_last(char *aString);

SchemaMapping::SchemaMapping(const string& schemaFile,
  const string& schemaFileOdb, bool verbose)
{

  Clear();

  // Schema file is required.
  if (schemaFile.empty() && schemaFileOdb.empty())
      return;

  _schemaFile = schemaFile;

  _schemaFileOdb = schemaFileOdb;

  //
  //----------------------------------------------------------------------
  //              Read schema and map data file ... 
  //

  if (_schemaFileOdb.empty())
  {
      _fobjS = new CifFile(_verbose, StringCompare::eCASE_SENSITIVE,
        _MAX_LINE_LENGTH);
  }
  else if (!_schemaFile.empty())
  {
      _fobjS = new CifFile(CREATE_MODE, _schemaFileOdb, _verbose,
        StringCompare::eCASE_SENSITIVE, _MAX_LINE_LENGTH);
  }
  else
  {
      _fobjS = new CifFile(READ_MODE, _schemaFileOdb, _verbose,
        StringCompare::eCASE_SENSITIVE, _MAX_LINE_LENGTH);
  }

  if (!_schemaFile.empty())
  {
      string diags;

      CifParser* cifParserR = new CifParser(_fobjS, _fobjS->GetVerbose());

      cifParserR->Parse(_schemaFile, diags);

      delete(cifParserR);

      if (!diags.empty())
      {
#ifdef VLAD_LOG_SEPARATION
        if (_verbose) _log << " Diagnostics [" << diags.size() << "] " << diags.c_str() << endl;
#endif
        diags.clear();
      }
  } 

  _AssignAttribIndices();

}


SchemaMapping::~SchemaMapping()
{

    if (_fobjS != NULL)
    {
        // Note that all tables that are in _fobjS will also be deleted here.
        delete(_fobjS);
    }

    delete(_schemaMap);

}


void SchemaMapping::Clear()
{

    _schemaFile.clear();
    _schemaFileOdb.clear();

    _fobjS = NULL;

    _tableSchema = NULL;
    _attribSchema = NULL;
    _schemaMap = NULL;
    _mapConditions = NULL;
    _tableAbbrev = NULL;
    _attributeAbbrev = NULL;
    _reviseSchemaMode  = false;

    _verbose = false;

}


void SchemaMapping::GetTableNames(vector<string>& tableNames)
{

    _tableSchema->GetColumn(tableNames, "table_name");

}


void SchemaMapping::GetGroupNames(vector<string>& groupNames)
{

    _tableSchema->GetColumn(groupNames, "group_name");

}


void SchemaMapping::GetAttributeNames(vector<string>& attributeNames, 
  const string& tableName)
{

    attributeNames.clear();

    vector<unsigned int> ir;

    _attribSchema->Search(ir, tableName, "table_name");
    if (ir.empty())
        return;

    _attribSchema->GetColumn(attributeNames, "attribute_name", ir);

}


void SchemaMapping::GetAttributesInfo(vector<AttrInfo>& attrInfo,
      const string& tableName)
{

    attrInfo.clear();

    vector<unsigned int> ir;

    _attribSchema->Search(ir, tableName, "table_name");
    if (ir.empty())
        return;

    for (unsigned int i = 0; i < ir.size(); ++i)
    {
        AttrInfo tmpAttrInfo;
 
        tmpAttrInfo.attribName = (*_attribSchema)(ir[i], "attribute_name");
        tmpAttrInfo.dataType = (*_attribSchema)(ir[i], "data_type");
        tmpAttrInfo.indexFlag = (*_attribSchema)(ir[i], "index_flag");
        tmpAttrInfo.nullFlag = (*_attribSchema)(ir[i], "null_flag");
        tmpAttrInfo.width = (*_attribSchema)(ir[i], "width");
        tmpAttrInfo.precision = (*_attribSchema)(ir[i], "precision");
        tmpAttrInfo.populated = (*_attribSchema)(ir[i], "populated");

        attrInfo.push_back(tmpAttrInfo);
        
    }

}


void SchemaMapping::GetMappedAttributesInfo(
  vector<vector<string> >& mappedAttrInfo, const string& tableName)
{

    mappedAttrInfo.clear();
 
    vector<unsigned int> irMap;

    // Find indices in attribute map of mapped attributes of the table
    _schemaMap->Search(irMap, tableName, "target_table_name");
    if (irMap.empty())
        return;

    vector<string> tmpVector;
    mappedAttrInfo.push_back(tmpVector);
    mappedAttrInfo.push_back(tmpVector);
    mappedAttrInfo.push_back(tmpVector);
    mappedAttrInfo.push_back(tmpVector);

    // Get mapped attribute names of this table
    _schemaMap->GetColumn(mappedAttrInfo[0], "target_attribute_name", irMap);

    // Get source CIF item names of the mapped attribute names of this table
    _schemaMap->GetColumn(mappedAttrInfo[1], "source_item_name", irMap);

    // Get condition id and function id of the mapped attribute names of
    //  this table
    _schemaMap->GetColumn(mappedAttrInfo[2], "condition_id", irMap);
    _schemaMap->GetColumn(mappedAttrInfo[3], "function_id", irMap);

}


void SchemaMapping::GetMappedConditions(
  vector<vector<string> >& mappedConditions, const string& tableName)
{

    mappedConditions.clear();
 
    vector<unsigned int> ir;

    _mapConditions->Search(ir, tableName, "condition_id");

    if (ir.empty())
        return;

    vector<string> tmpVector;
    mappedConditions.push_back(tmpVector);
    mappedConditions.push_back(tmpVector);

    _mapConditions->GetColumn(mappedConditions[0], "attribute_name", ir);
    _mapConditions->GetColumn(mappedConditions[1], "attribute_value", ir);

}


void SchemaMapping::ReviseSchemaMap(const string& mapFile)
{

    if (!_reviseSchemaMode || mapFile.empty())
        return;

    if (_fobjS)
    {
        cout << "Writing revised schema map file " << mapFile << endl;
        _fobjS->Write(mapFile);
    }

}


void SchemaMapping::SetReviseSchemaMode(bool mode)
{
    _reviseSchemaMode = mode;
}


bool SchemaMapping::GetReviseSchemaMode()
{
    return(_reviseSchemaMode);
}


void SchemaMapping::_AssignAttribIndices()
{

  unsigned int i;
  Block& rsBlock = _fobjS->GetBlock("rcsb_schema");
  Block& rsmBlock = _fobjS->GetBlock("rcsb_schema_map");

  _tableSchema     = rsBlock.GetTablePtr("rcsb_table");
  _attribSchema    = rsBlock.GetTablePtr("rcsb_attribute_def");
  _schemaMap       = rsmBlock.GetTablePtr("rcsb_attribute_map");
  _mapConditions   = rsmBlock.GetTablePtr("rcsb_map_condition");
  _tableAbbrev     = rsBlock.GetTablePtr("rcsb_table_abbrev");
  _attributeAbbrev = rsBlock.GetTablePtr("rcsb_attribute_abbrev");

  if ((!_tableAbbrev->IsColumnPresent("table_name")) || (!_tableAbbrev->IsColumnPresent("table_abbrev"))) {
#ifdef VLAD_LOG_SEPARATION
    if (_verbose) _log << "Table abbreviations incomplete" << endl;
#endif
    return;
  }

  if ((!_attributeAbbrev->IsColumnPresent("table_name")) || (!_attributeAbbrev->IsColumnPresent("attribute_abbrev")) || (!_attributeAbbrev->IsColumnPresent("attribute_abbrev"))) {
#ifdef VLAD_LOG_SEPARATION
    if (_verbose) _log << "Attribute abbreviations incomplete" << endl;
#endif
    return;
  }

  if ((!_tableSchema->IsColumnPresent("table_name")) || (!_tableSchema->IsColumnPresent("group_name"))) {
#ifdef VLAD_LOG_SEPARATION
    if (_verbose) _log << "Table schema table incomplete" << endl;
#endif
    return;
  }

  if (!_attribSchema->IsColumnPresent("populated"))
  {
    _attribSchema->AddColumn("populated");
  }
  // VLAD else .. = -1

  if ((!_attribSchema->IsColumnPresent("table_name")) || (!_attribSchema->IsColumnPresent("attribute_name")) || (!_attribSchema->IsColumnPresent("data_type")) || (!_attribSchema->IsColumnPresent("index_flag")) || (!_attribSchema->IsColumnPresent("null_flag")) || (!_attribSchema->IsColumnPresent("width")) || (!_attribSchema->IsColumnPresent("precision"))) {
#ifdef VLAD_LOG_SEPARATION
    if (_verbose) _log << "Attribute schema table incomplete" << endl;
#endif
    return;
  }

  if ((!_schemaMap->IsColumnPresent("target_table_name")) || (!_schemaMap->IsColumnPresent("target_attribute_name")) || 
	(!_schemaMap->IsColumnPresent("source_item_name")) || (!_schemaMap->IsColumnPresent("condition_id"))  || (!_schemaMap->IsColumnPresent("function_id"))) {
#ifdef VLAD_LOG_SEPARATION
    if (_verbose) _log << "Schema map table incomplete" << endl;
#endif
    return;
  }

  if ((!_mapConditions->IsColumnPresent("condition_id")) || (!_mapConditions->IsColumnPresent("attribute_name")) || (!_mapConditions->IsColumnPresent("attribute_value"))) {
#ifdef VLAD_LOG_SEPARATION
    if (_verbose) _log << "Map condition table incomplete" << endl;
#endif
    return;
  }

  //
  // Check that all schema attributes have a CIF mapping....
  //
  vector<unsigned int> ir;
  vector<string> qcol, qval;    
  qcol.clear(); qcol.push_back("target_table_name"); qcol.push_back("target_attribute_name");

  //
  string cs1, cs2, cs3;
  //

  ISTable * t = new ISTable("rcsb_attribute_map1");
  //  vector<string> *trow = NULL;
  string empty("");
  t->AddColumn("target_table_name");
  t->AddColumn("target_attribute_name");
  t->AddColumn("source_item_name");
  t->AddColumn("condition_id");
  t->AddColumn("function_id");
  
  qcol.clear(); qcol.push_back("table_name"); qcol.push_back("attribute_name");

  for (i=0; i < _schemaMap->GetNumRows(); i++) {
    const vector<string>& row = _schemaMap->GetRow(i);
    if (!row.empty()) {
      qval.clear();
      qval.push_back((*_schemaMap)(i, "target_table_name"));
      qval.push_back((*_schemaMap)(i, "target_attribute_name"));

    } else {
      continue;
    }
#if 0
    // JDW test 
    cs1.clear(); cs1 = (*_schemaMap)(i, "target_table_name");
    cs2.clear(); cs2 = (*_schemaMap)(i, "target_attribute_name");
    _db._schemaMapping.GetTableNameAbbrev(cs3, cs1);
    cerr << "table " << cs1.c_str() << " abbrev " << cs3.c_str() << endl;

    _db._schemaMapping.GetAttributeNameAbbrev(cs3, cs1, cs2);
    cerr << "table " << cs1.c_str() << " attrib " << cs2.c_str() << " abbrev " << cs3.c_str() << endl;
#endif

    //
    _attribSchema->Search(ir, qval, qcol);
    if (!ir.empty()) {
      t->AddRow(row);
      ir.clear();
      continue;
    } else {
      //      cerr << "In " << _INPUT_FILE << ": " << "No schema correspondence for " << "_"<< qval[0].c_str() 
      //	   << "."  <<qval[1].c_str() << endl;
#ifdef VLAD_LOG_SEPARATION
      if (_verbose) _log << "No schema correspondence for " 
			 << ((*_schemaMap)(i, "source_item_name")).c_str() 
			 << endl;
#endif
    }

  }
  _schemaMap = t;
}


void SchemaMapping::GetTableNameAbbrev(string &abbrev, string target)
{

  vector<unsigned int> ir;
  vector<string> qcol, qval;    

  abbrev.clear(); abbrev = target;
  
  qcol.clear(); qcol.push_back("table_name");
  qval.clear(); qval.push_back(target);
  _tableAbbrev->Search(ir, qval, qcol);

  if (ir.empty()) return;
  abbrev.clear();
  abbrev = (*_tableAbbrev)(ir[0], "table_abbrev");
  ir.clear();

}

void SchemaMapping::GetAttributeNameAbbrev(string &abbrev, string targetTable, string targetAttribute)
{

  vector<unsigned int> ir;

  vector<string> qcol, qval;    
  abbrev.clear(); abbrev = targetAttribute;

  qcol.clear(); qcol.push_back("table_name"); qcol.push_back("attribute_name");
  qval.clear(); qval.push_back(targetTable); qval.push_back(targetAttribute);
  _attributeAbbrev->Search(ir, qval, qcol);

  if (ir.empty()) return;
  abbrev.clear();
  abbrev = (*_attributeAbbrev)(ir[0], "attribute_abbrev");
  ir.clear();

}


void SchemaMapping::getSchemaMapDetails(const string& tableName,
  const string& attribName, string& dataType, string& indexFlag,
  string& nullFlag, string& width, string &precision, string& populated)
{

    dataType.clear();
    indexFlag.clear();
    nullFlag.clear();
    width.clear();
    precision.clear();
    populated.clear();

    //
    // Get attibute schema details... 
    //
    vector<unsigned int> ir;
    vector<string> qcol, qval;    

    qcol.push_back("table_name");
    qcol.push_back("attribute_name");

    qval.push_back(tableName);
    qval.push_back(attribName);

    _attribSchema->Search(ir, qval, qcol);

    dataType = (*_attribSchema)(ir[0], "data_type");
    indexFlag = (*_attribSchema)(ir[0], "index_flag");
    nullFlag = (*_attribSchema)(ir[0], "null_flag");
    width = (*_attribSchema)(ir[0], "width");
    precision = (*_attribSchema)(ir[0], "precision");
    populated = (*_attribSchema)(ir[0], "populated");

}


void SchemaMapping::updateSchemaMapDetails(SchemaMapping& scm)
{

    unsigned int numPopulated = 0;
    unsigned int numPopulatedR = 0;
    unsigned int numPopulatedU = 0;

    unsigned int nRows = _attribSchema->GetNumRows();

    for (unsigned int i = 0; i < nRows; ++i)
    {
        const string& attribName = (*_attribSchema)(i, "attribute_name");
        const string& tableName = (*_attribSchema)(i, "table_name");

        const string& width = (*_attribSchema)(i, "width");
        const string& populated = (*_attribSchema)(i, "populated");

        int iWidth = atoi(width.c_str());

        string dataTypeU, indexFlagU, nullFlagU, widthU, precisionU,
          populatedU;

        scm.getSchemaMapDetails(tableName, attribName, dataTypeU, indexFlagU,
          nullFlagU, widthU, precisionU, populatedU);

        int iWidthU = atoi(widthU.c_str());

        if (populated == "Y")
            numPopulated++;

        if (populatedU == "Y")
            numPopulatedR++;

        if (populated == "Y" || populatedU == "Y")
        {
            _set_attribute_value_where(_attribSchema, tableName, attribName,
              "populated", "Y");
            numPopulatedU++;
        }
        else
        {
            _set_attribute_value_where(_attribSchema, tableName, attribName,
              "populated", "N");
        }

        if (iWidthU > iWidth)
        {
            _set_attribute_value_where(_attribSchema, tableName, attribName,
              "width", widthU);
        }
    }

    cout << "Items populated in the reference map = " << numPopulated << endl;
    cout << "Items populated in the revised   map = " << numPopulatedR << endl;
    cout << "Items populated in the updated   map = " << numPopulatedU << endl;

}


void SchemaMapping::UpdateAttributeDef(const string& tableName,
  const string& columnName, int iType, int iWidth, int newWidth)
{
  if ((iType == SchemaMapping::_TYPE_CODE_STRING || iType == SchemaMapping::_TYPE_CODE_TEXT ) && (newWidth > iWidth)) {
#ifdef VLAD_LOG_SEPARATION
    if (_verbose) _log << "Updating max width for " << tableName.c_str() << " " << columnName << " to " << newWidth << endl;
#endif
    if (newWidth >= 10 && newWidth < 80) {
      newWidth = newWidth+1;
    } else if (newWidth >= 80 && newWidth < 128) {
      newWidth = 128;
    } else if (newWidth >= 128 && newWidth < 255) {
      newWidth = 255;
    } else if (newWidth >= 255 && newWidth < 511) {
      newWidth = 511;
    } else if (newWidth >= 512 && newWidth < 1023) {
      newWidth = 1023;
    } else if (newWidth >= 1024 && newWidth < 4095) {
      newWidth = 4095;
    } else if (newWidth >= 4096 && newWidth < 16383) {
      newWidth = 16383;
    } else if (newWidth >= 16384 && newWidth < 32000) {
      newWidth = 32000;
    } else if (newWidth >= 32001 && newWidth < 64000) {
      newWidth = 64000;
    }

    string width = String::IntToString(newWidth);

    _set_attribute_value_where(_attribSchema, tableName, columnName, "width", width);
    
    //    if ((_serverType == _SERVER_TYPE_MYSQL || _serverType == _SERVER_TYPE_SYBASE) && newWidth > 255) 
    //      _set_attribute_value_where(_attribSchema, tableName, columnName, "data_type", "text");
  }     
  _set_attribute_value_where(_attribSchema, tableName, columnName, "populated", "Y");
}


Db::Db(SchemaMapping& schemaMapping, const string& dbName) :
  _schemaMapping(schemaMapping), _useOnlyPopulated(false), _appendFlag(false),
  _dbName(dbName)
{

  _fieldSeparator.push_back('\t');

  _rowSeparator.push_back('\n');

  _dataLoadingFileName = _SCRIPT_LOADING_FILE;

}


Db::~Db()
{

}


void Db::SetAppendFlag(const bool appendFlag)
{
    _appendFlag = appendFlag;
}


bool Db::GetAppendFlag()
{

    return(_appendFlag);

}


bool DbOutput::IsFirstTextNewLineSpecial()
{

    return(false);

}


const string& DbOutput::GetCommandScriptName()
{

    return(_DATA_LOADING_SCRIPT);

}


bool SqlOutput::IsFirstTextNewLineSpecial()
{

    return(_db.IsFirstTextNewLineSpecial());

}


bool Db::IsFirstTextNewLineSpecial()
{

    return(_firstTextNewLineSpecial);

}

void Db::SetFieldSeparator(const string& fieldSeparator)
{

    if (fieldSeparator.empty())
        return;

    _fieldSeparator = fieldSeparator;

}

void Db::SetRowSeparator(const string& rowSeparator)
{

    if (rowSeparator.empty())
        return;

    _rowSeparator = rowSeparator;

}


string Db::GetCommandTerm()
{

    return(_cmdTerm);

}

string Db::GetFieldSeparator()
{

    return(_fieldSeparator);

}


string Db::GetRowSeparator()
{

    return(_rowSeparator);

}


void DbMySql::GetStart(string& start)
{

    start = "USE " + _dbName + _cmdTerm + '\n';

}


void DbSybase::GetStart(string& start)
{

    start = "USE " + _dbName + _cmdTerm + '\n';

}


void Db::WriteSchemaStart(ostream& io)
{

    io << "USE " << _dbName << _cmdTerm << endl;

}


void Db::WriteDeleteTable(ostream& io, const string& fromTable,
  const string& where, const string& what)
{

    io << "DELETE FROM " << fromTable << " WHERE " << where << " = '"  <<
      what << "'" << _cmdTerm << endl;

}


void SqlOutput::WriteSqlScriptSchemaInfo(ostream& io)
{

    // VLAD - IMPROVE
    // WriteScriptSchemaInfo can be created and used by BcpOutput.
 
    WriteHeader(io);
    io << endl;

    io << "if (\"x$1\" == \"xdrop\") then " << endl;

    WriteDbExec(io, _SCHEMA_DELETE_FILE, 1);

    io << "endif" << endl;
    io << endl;

    WriteDbExec(io, _SCHEMA_FILE);

    io << endl;

}


void DbOutput::WriteDbExec(ostream& io, const string& fileName,
  const unsigned int indentLevel)
{

    for (unsigned int i = 0; i < indentLevel; ++i)
    {
        io << "    ";
    }

    io << "if (-e " << fileName << " && ! -z " << fileName << ") then" << endl;

    WriteDbExecOnly(io, fileName, indentLevel + 1);

    for (unsigned int i = 0; i < indentLevel; ++i)
    {
        io << "    ";
    }

    io << "endif" << endl;

}


void DbOutput::WriteDbExecOnly(ostream& io, const string& fileName,
  const unsigned int indentLevel)
{

    const string& connect = _db.GetConnect();

    if (!connect.empty())
    {
        for (unsigned int i = 0; i < indentLevel; ++i)
        {
            io << "    ";
        }

        io << connect << endl;

    }

    for (unsigned int i = 0; i < indentLevel; ++i)
    {
        io << "    ";
    }

    io << _db.GetDbCommand() << " " << fileName << endl;

    const string& terminate = _db.GetTerminate();

    if (!terminate.empty())
    {
        for (unsigned int i = 0; i < indentLevel; ++i)
        {
            io << "    ";
        }

        io << terminate << endl;
    }

}


void Db::DropTableSql(ostream& io, const string& tableNameDb)
{

    io << "DROP TABLE " << tableNameDb << _cmdTerm << endl;

}


const string& Db::GetExec()
{

    return(_exec);

}


const string& Db::GetExecOption()
{

    return(_execOption);

}


const string& Db::GetUserOption()
{

    return(_userOption);

}


const string& Db::GetPassOption()
{

    return(_passOption);

}


const string& Db::GetEnvDbUser()
{

    return(_envDbUser);

}


const string& Db::GetEnvDbPass()
{

    return(_envDbPass);

}


const string& Db::GetDataLoadingFileName()
{

    return(_dataLoadingFileName);

}


const string& Db::GetBcpStringDelimiter()
{

    return(_bcpStringDelimiter);

}


void Db::WriteLoadingStart(ostream& io)
{


}


void Db::WriteLoadingEnd(ostream& io)
{


}


void Db::WriteLoadingTable(ostream& io, const string& tableName,
  const string& workDir)
{


}


void Db::WritePrint(ostream& io, const string& tableNameDb)
{

}


void DbSybase::WritePrint(ostream& io, const string& tableNameDb)
{

    io << "PRINT \"Creating table " << tableNameDb << "\"" << endl;

}


void Db::GetChar(string& dType, const unsigned int width)
{

    dType = "varchar(";
    dType += String::IntToString(width);
    dType += ")";

}


void Db::GetText(string& dType, const unsigned int width)
{


}


void DbOracle::GetText(string& dType, const unsigned int width)
{

    if (width > 4000)
    {
        dType = "varchar(4000)";
    }
    else
    {
        dType = "clob";
    }

}


void DbDb2::GetText(string& dType, const unsigned int width)
{

    if (width == 0 )
    {
        dType = "varchar(4000)";
    }
    else if (width < 32000 )
    {
        dType = "varchar(";
        dType += String::IntToString(width);
        dType += ")";
    }
    else
    {
        dType = "clob(";
        dType += String::IntToString(width);
        dType += ")";
    }

}


void Db::GetFloat(string& dType)
{

}


void Db::GetDate(string& dType)
{

}

void Db::WriteNull(ostream& io, const int iNull,
  const unsigned int curr, const unsigned int attSize)
{

}


void Db::WriteBcpDoubleQuotes(ostream& io)
{

    io << "\"";

}


void DbOutput::_FormatData(ostream &io, const string &cs,
  ATTRIBINFO& attribInfo, const bool noRevise)
{

    if (!noRevise)
    {
        unsigned int jLen = cs.size();
        if (jLen > attribInfo.iMaxWidth)
            attribInfo.iMaxWidth = jLen;
    }

    switch (attribInfo.iTypeCode)
    {
        case SchemaMapping::_TYPE_CODE_INT: 
        case SchemaMapping::_TYPE_CODE_FLOAT:
        case SchemaMapping::_TYPE_CODE_BIGINT:
        {
            _FormatNumericData(io, cs);
            break;
        }
        case SchemaMapping::_TYPE_CODE_STRING:
        {
            if (!noRevise && _db._schemaMapping.GetReviseSchemaMode())
            {
                _FormatStringData(io, cs, attribInfo.iMaxWidth);
            }
            else
            {
                _FormatStringData(io, cs, attribInfo.iWidth);
            }
            break;
        }
        case SchemaMapping::_TYPE_CODE_TEXT:
        {
            _FormatTextData(io, cs);
            break;
        }
        case SchemaMapping::_TYPE_CODE_DATETIME:
        {
            _FormatDateData(io, cs, attribInfo.iWidth);
            break;
        }
    }

}


void DbOutput::WriteNewLine(ostream& io, bool special)
{

    io << "\n";

}


void DbOutput::GetTableStart(string& tableStart, const string& tableName)
{

    tableStart.clear();

}


void DbOutput::GetTableEnd(string& tableEnd)
{

    tableEnd.clear();

}


void SqlOutput::GetTableStart(string& tableStart, const string& tableName)
{

    tableStart.clear();

    if (tableName.empty())
    {
        return;
    }

    string tableNameDb;
    _db._schemaMapping.GetTableNameAbbrev(tableNameDb, tableName);

    tableStart += "INSERT INTO ";
    tableStart += tableNameDb;
    tableStart += "\n";
    tableStart += "VALUES (";
 
}


void SqlOutput::GetTableEnd(string& tableEnd)
{

    tableEnd.clear();

    tableEnd += "\n";
    tableEnd += ")";
    tableEnd += _db.GetCommandTerm();
    tableEnd += "\n\n";

}


const string& DbOutput::GetItemSeparator()
{

    return(_itemSeparator);

}


const string& DbOutput::GetRowSeparator()
{

    return(_rowSeparator);

}


void DbOutput::_WriteTable(ostream& io, ISTable* tIn, bool noRevise)
{

    if (!io || !tIn)
        return;

    string tableStart;
    GetTableStart(tableStart, tIn->GetName());

#ifdef DEBUG_POINT
    if (tIn->GetName() == "citation")
    {
        int a = 1;
        int b = a + 3;
        ++b;
    }
#endif

    string tableEnd;
    GetTableEnd(tableEnd);

    // Get all of the attributes for this table.
    vector<ATTRIBINFO> aI;
    _db._schemaMapping.GetAttributeInfo(aI, tIn);

    unsigned int nRows = tIn->GetNumRows();
    unsigned int nCols = tIn->GetNumColumns();

    for (unsigned int i = 0; i < nRows; ++i)
    {
        const vector<string>& row = tIn->GetRow(i);

        if (row.empty())
        {
            continue;
        }

        if (!SchemaMapping::AreAttributesMandatory(row, aI))
        {
            continue;
        }

        io << tableStart;

        for (unsigned int j = 0; j < nCols; ++j)
        {
	    if (j != 0)
                io << GetItemSeparator();

            _FormatData(io, row[j], aI[j], noRevise);
        }

        io << GetRowSeparator();

        io << tableEnd;
    }

    if (noRevise)
    {
        return;
    }

    const vector<string>& columnNames = tIn->GetColumnNames();

#ifdef VLAD_LOG_SEPARATION
    if (_verbose)
    {
        for (unsigned int i = 0; i < nCols; ++i)
        {
            _log  << tIn->GetName() << " " << columnNames[i] << " " <<
              aI[i].iMaxWidth << endl;
        }
    }
#endif

    if (_db._schemaMapping.GetReviseSchemaMode())
    {
        for (unsigned int i = 0; i < nCols; ++i)
        {
            _db._schemaMapping.UpdateAttributeDef(tIn->GetName(),
              columnNames[i], aI[i].iTypeCode, aI[i].iWidth, aI[i].iMaxWidth);
        }
    }

}


void SqlOutput::WriteNewLine(ostream& io, bool special)
{

    _db.WriteNewLine(io, special);

}


void Db::WriteNewLine(ostream& io, bool special)
{

    io << "\n";

}


void DbMySql::WriteNewLine(ostream& io, bool special)
{

    io << "\\n";

}


void DbOracle::WriteNewLine(ostream& io, bool special)
{

    if (special)
        io << " \\" << endl;
    else
        io << "\\" << endl;

}


void DbDb2::WriteBcpDoubleQuotes(ostream& io)
{

    io << "\"\"";

}


void DbDb2::WriteNull(ostream& io, const int iNull,
  const unsigned int curr, const unsigned int attSize)
{

    if (iNull)
    {
        io << "not null";
    }
    else
    {
        io << "        ";
    }

    io << "," << endl;

}


void DbOracle::WriteNull(ostream& io, const int iNull,
  const unsigned int curr, const unsigned int attSize)
{

    if (iNull)
    {
        io << "not null";
    }
    else
    {
        io << "    null";
    }

    io << "," << endl;

}


void DbMySql::WriteNull(ostream& io, const int iNull, const unsigned int curr,
  const unsigned int attSize)
{

    if (iNull)
    {
        io << "not null";
    }
    else
    {
        io << "    null";
    }

    if (curr < attSize - 1)
        io << "," << endl;
    else
        io << endl;

}


void DbSybase::WriteNull(ostream& io, const int iNull,
  const unsigned int curr, const unsigned int attSize)
{

    if (iNull)
    {
        io << "not null";
    }
    else
    {
        io << "    null";
    }

    if (curr < attSize - 1)
        io << "," << endl;
    else
        io << endl;

}


void Db::WriteTableIndex(ostream& io, const string& tableNameDb,
      const vector<string>& indexList)
{

}


void DbOracle::WriteTableIndex(ostream& io, const string& tableNameDb,
      const vector<string>& indexList)
{

    io << "PRIMARY KEY (";

    for (unsigned int i = 0; i < indexList.size(); ++i)
    {
        io << indexList[i];

        if (i < indexList.size() - 1)
            io << ",";
        else
            io << " ";
    }

    io << ")" << endl;

    io << ")" <<  _cmdTerm << endl;  // end of create table clause

}


void DbDb2::WriteTableIndex(ostream& io, const string& tableNameDb,
      const vector<string>& indexList)
{

    io << "PRIMARY KEY (";
    for (unsigned int i = 0; i < indexList.size(); ++i)
    {
        io << indexList[i];

        if (i < indexList.size() - 1)
            io << ",";
        else
            io << " ";
    }

    io << ")" << endl;

    io << ")" <<  _cmdTerm << endl;  // end of create table clause


    if (!indexList.empty())
    {
        io << _cmdTerm << endl;
    }

}


void DbMySql::WriteTableIndex(ostream& io, const string& tableNameDb,
      const vector<string>& indexList)
{

    io << ")" <<  _cmdTerm << endl << endl;  // end of create table clause

    if (!indexList.empty())
    {
        io << "CREATE UNIQUE INDEX primary_index ON " << tableNameDb << endl;
        io << "(" << endl;
        unsigned int indLen  = indexList.size();
        if (indLen > 16) indLen = 16;
        for (unsigned int i = 0; i < indLen; ++i)
        {
            io << indexList[i];
            if (i < indLen - 1)
                io << "," << endl;
            else
                io << endl;
        }

        io << ")";

        io << _cmdTerm << endl;
    }

}


void DbSybase::WriteTableIndex(ostream& io, const string& tableNameDb,
      const vector<string>& indexList)
{

    io << ")" <<  _cmdTerm << endl << endl;  // end of create table clause


    if (!indexList.empty())
    {
        io << "CREATE CLUSTERED INDEX primary_index ON " << tableNameDb <<
          endl;
        io << "(" << endl;
        for (unsigned int i = 0; i < indexList.size(); ++i)
        {
            io << indexList[i];
            if (i < indexList.size() - 1)
                io << "," << endl;
            else
                io << endl;
        }
        io << ")";

        io << " WITH ignore_dup_row" << _cmdTerm << endl;
    }

}


void DbOracle::GetDate(string& dType)
{

    dType = "date";

}


void DbDb2::GetDate(string& dType)
{

    dType = "date";

}


void DbDb2::GetFloat(string& dType)
{

    dType = "decimal(14,4)";

}


const string& Db::GetConnect()
{

    return(_connect);

}


const string& Db::GetTerminate()
{

    return(_terminate);

}


const string& Db::GetDbCommand()
{

    return(_dbCommand);

}


DbOracle::DbOracle(SchemaMapping& schemaMapping, const string& dbName) :
  Db(schemaMapping, dbName)
{

    _cmdTerm.push_back(';');

    _dbCommand = "$ORACLE_HOME/bin/sqlplus $dbuser/$dbpw@" + _dbName + " <";

    _envDbUser = "NDB_XDBUSER";
    _envDbPass = "NDB_XDBPW";

    _firstTextNewLineSpecial = true;

}


DbOracle::~DbOracle()
{

}


void DbOracle::WriteLoadingStart(ostream& io)
{

    io << "#!/bin/csh -f" << endl << endl;

    io << "set dbuser = $" << GetEnvDbUser() << endl;
    io << "set dbpw = $" << GetEnvDbPass() << endl;

    io << endl;

}


void DbOracle::WriteLoadingTable(ostream& io, const string& tableName,
  const string& workDir)
{

    string fs;
    escapeString(fs, _fieldSeparator);

    string rs;
    escapeString(rs, _rowSeparator);

        string tableNameDb;
        _schemaMapping.GetTableNameAbbrev(tableNameDb, tableName);

        vector<string> colNames;
        _schemaMapping.GetAttributeNames(colNames, tableName);
        if (colNames.empty())
        {
            return;
        }

        unsigned int nCols = colNames.size();

        vector<ATTRIBINFO> aI;
        _schemaMapping.GetAllAttributeInfo(aI, tableName);

        bool iText = false;
        for (unsigned int j = 0; j < nCols - 1; ++j)
        {
	    if (aI[j].iTypeCode == SchemaMapping::_TYPE_CODE_TEXT)
            {
	        iText = true; 
                break;
	    }
        }

        string directOption;

        if (iText)
            directOption = "false";
        else
            directOption = "true";

        string extension = ".ctl";

        for (unsigned int filesI = 0; filesI < 4; ++filesI, extension += "+")
        {
            io << "if (-e " << tableName << extension << " && ! -z " <<
              tableName << extension << ") then" << endl;
            io << "   echo \"Loading table " << tableName << "\" "<< endl;

            io << "   $ORACLE_HOME/bin/sqlldr $dbuser/$dbpw@" << _dbName <<
              " " << tableName << extension << " direct=" << directOption <<
              endl;

            io << "endif" << endl;

            io << endl;
        }

        struct stat statbuf;

        string tName = workDir;
        tName += tableName;
        tName += ".bcp";
        int istat = stat(tName.c_str(), &statbuf);

        string oFile = workDir;
        oFile += tableName;
        oFile += ".ctl";

        string bFile = workDir;
        bFile += tableName;
        bFile += ".bad";

        bool ifirst = true;

        while (ifirst || (istat == 0 && statbuf.st_size > 0))
        {
#ifdef VLAD_LOG_SEPARATION
            if (_verbose) _log << "Writing control file for " << tName << endl;
#endif
            ofstream ioctl(oFile.c_str(), ios::out | ios::trunc);

            ifirst = false;

            ioctl << "OPTIONS ( ERRORS=100000 )" << endl;
            ioctl << "LOAD DATA" << endl;
            ioctl << "INFILE '" << tName << "'  \"str '" << rs << "'\"" << endl;
            ioctl << "BADFILE '" << bFile <<"'" << endl;
            ioctl << "DISCARDMAX 100000"  << endl;
      
            if (_appendFlag)
                ioctl << "APPEND" << endl;
            else 
	        ioctl << "TRUNCATE" << endl;
      
            ioctl << "INTO TABLE " << tableNameDb << endl;
            ioctl << "FIELDS TERMINATED BY '" << fs << "'" << endl;
            ioctl << "TRAILING NULLCOLS" << endl;
      
            ioctl << "(";
      
            for (unsigned int j = 0; j < nCols; ++j)
            {
                string columnNameDb;
	        _schemaMapping.GetAttributeNameAbbrev(columnNameDb,
                  tableName, colNames[j]);

	        switch (aI[j].iTypeCode)
                {
	            case SchemaMapping::_TYPE_CODE_INT: 
	            case SchemaMapping::_TYPE_CODE_FLOAT:
	            case SchemaMapping::_TYPE_CODE_BIGINT:
	                ioctl <<  columnNameDb;
	                break;
                    case SchemaMapping::_TYPE_CODE_STRING:
                        ioctl <<  columnNameDb << " char(" <<
                          aI[j].iWidth << ")";
                        break;
                    case SchemaMapping::_TYPE_CODE_TEXT:
                        ioctl << columnNameDb << " char(100000)";
                        break;
                    case SchemaMapping::_TYPE_CODE_DATETIME:
                        ioctl << columnNameDb <<
                          " date(30) \"YYYY-MM-DD HH24:MI:SS\"";
                        break;
                }

	        if (j != nCols-1)
                    ioctl << "," << endl;
            }
            ioctl << ")" << endl << endl;
            ioctl.close();

            // +++++++++++++++++
           tName += "+";
           istat = stat(tName.c_str(),&statbuf);
           oFile += "+";
           bFile += "+";
       }

}


void Db::GetStart(string& start)
{

    start.clear();

}


void DbOracle::WriteSchemaStart(ostream& io)
{

    io << "SET DEFINE OFF" << _cmdTerm << endl;
    io << "SET ESCAPE \\" << _cmdTerm << endl;

}


DbDb2::DbDb2(SchemaMapping& schemaMapping, const string& dbName) :
  Db(schemaMapping, dbName)
{

    _cmdTerm.push_back(';');

    _connect = "db2 CONNECT TO " + _dbName + " USER $dbuser USING $dbpw";
    _terminate = "db2 TERMINATE ";
    _dbCommand = "db2 -tf";
    _envDbUser = "DB2_DB_USER";
    _envDbPass = "DB2_DB_PW";

    _bcpStringDelimiter = "\"";

}


DbDb2::~DbDb2()
{

}


void DbDb2::GetStart(string& start)
{

    start = "CONNECT TO " + _dbName + " USER $dbuser USING $dbpw" +
      _cmdTerm + '\n';

}


void DbDb2::WriteSchemaStart(ostream& io)
{

    string schemaStart;
    GetStart(schemaStart);

    io << schemaStart;

}


void SqlOutput::CreateTableSql(ostream& io, const string& tableName,
  const vector<AttrInfo>& attrInfo)
{

    if (!io)
        return;

    string tableNameDb;
    _db._schemaMapping.GetTableNameAbbrev(tableNameDb, tableName);

    _db.WritePrint(io, tableNameDb);

    io << "CREATE TABLE " << tableNameDb << endl;
    io << "(" << endl;

    vector<string> indexList;

    for (unsigned int i = 0; i < attrInfo.size(); ++i)
    {
        if (_db.GetUseOnlyPopulated() && (attrInfo[i].populated != "Y"))
        {
            continue;
        }

        const string& columnName = attrInfo[i].attribName;

        string columnNameDb;
        _db._schemaMapping.GetAttributeNameAbbrev(columnNameDb,
          tableName,columnName);

        string dType = attrInfo[i].dataType;
        int iIndex     = atoi((attrInfo[i].indexFlag).c_str());
        int iNull      = atoi((attrInfo[i].nullFlag).c_str());
        int iWidth     = atoi((attrInfo[i].width).c_str());
        //      iPrecision = atoi((attrInfo[i].precision).c_str());
        //JDW
        //     io << columnName.c_str();

        unsigned int iLen = columnNameDb.size();
        if (iLen > _MAX_SQL_NAME_LENGTH)
            iLen = _MAX_SQL_NAME_LENGTH;
        for (unsigned int j = 0; j < iLen; ++j)
        {
            io << columnNameDb[j];
        }

        for (unsigned int j = iLen; j < _MAX_SQL_NAME_LENGTH+1; ++j)
        {
            io << " ";
        }

        if (String::IsCiEqual(dType, "char"))
        {
            _db.GetChar(dType, iWidth);
        }

        if (String::IsCiEqual(dType, "text"))
        {
            _db.GetText(dType, iWidth);
        }

        if (String::IsCiEqual(dType, "datetime"))
        {
            _db.GetDate(dType);
        }

        if (String::IsCiEqual(dType, "float"))
        {
            _db.GetFloat(dType);
        }

        io << dType;
        for (unsigned int j = dType.size(); j < 15; ++j)
        {
            io << " ";
        }

        _db.WriteNull(io, iNull, i, attrInfo.size());

        if (iIndex)
            indexList.push_back(columnNameDb);
    }

    io << endl;

    _db.WriteTableIndex(io, tableNameDb, indexList);

}

void DbDb2::WriteLoadingStart(ostream& io)
{

    io << "#!/bin/csh -f" << endl << endl;

    io << "set dbuser = $" << GetEnvDbUser() << endl;
    io << "set dbpw = $" << GetEnvDbPass() << endl;

    io << endl;

    io << _connect << endl;

    io << endl;

}


void DbDb2::WriteLoadingEnd(ostream& io)
{

    io << endl;
    io << _terminate << endl;

}


void DbDb2::WriteLoadingTable(ostream& io, const string& tableName,
  const string& workDir)
{

        string tableNameDb;

        _schemaMapping.GetTableNameAbbrev(tableNameDb, tableName);

#ifdef VLAD_LOG_SEPARATION
        if (_verbose) _log  << "Writing DB2 load commands for table "
          <<  tableNameDb << endl;
#endif

        string tName = workDir;
        tName += tableName;
        tName += ".bcp";

        string mFile = workDir;
        mFile += tableName;
        mFile += ".msg";

        struct stat statbuf;

        int istat = stat(tName.c_str(), &statbuf);
        while (istat == 0 && statbuf.st_size > 0)
        {
            io << "echo \"Loading data from file  " << tName << "\"" << endl; 
            io << "db2 \"load from " << tName <<
              " of del modified by delprioritychar anyorder ";
            io << " messages " << mFile << " insert into ";
            io << tableNameDb << "\"" <<  endl;
            io << endl;
      
            // +++++++++++++++++
            tName += "+";
            istat = stat(tName.c_str(), &statbuf);
        }

}


DbMySql::DbMySql(SchemaMapping& schemaMapping, const string& dbName) :
  Db(schemaMapping, dbName)
{

    _cmdTerm.push_back(';');

    _exec = "mysql";
    _execOption = "-f ";
    _userOption = "--user=";
    _passOption = "--password=";
 
    _dbCommand = _exec + " " + _execOption + _userOption +
      "$dbuser" + " " + _passOption + "$dbpw <";

    _envDbUser = "NDB_XDBUSER";
    _envDbPass = "NDB_XDBPW";

    _dataLoadingFileName = _SQL_LOADING_FILE;
}


DbMySql::~DbMySql()
{

}


void DbMySql::DropTableSql(ostream& io, const string& tableNameDb)
{

    io << "DROP TABLE IF EXISTS " << tableNameDb << _cmdTerm << endl;

}


void DbMySql::WriteLoad(ostream& io)
{

    io << "mysql -v -f --user=$dbuser --password=$dbpw < " <<
      GetDataLoadingFileName()  << endl;

}


void DbMySql::WriteLoadingStart(ostream& io)
{

    io << "USE " << _dbName << _cmdTerm << endl;

}


void DbMySql::WriteLoadingTable(ostream& io, const string& tableName,
  const string& workDir)
{
    string fs;
    escapeString(fs, _fieldSeparator);

    string rs;
    escapeString(rs, _rowSeparator);


        string tName = workDir;
        tName += tableName;
        tName += ".bcp";

        struct stat statbuf;
        int istat = stat(tName.c_str(), &statbuf);
        while (istat == 0 && statbuf.st_size > 0)
        {
            string tableNameDb;
            _schemaMapping.GetTableNameAbbrev(tableNameDb, tableName);

            io << "LOAD DATA LOCAL INFILE \"" << tName << "\"" << endl
              << "IGNORE INTO TABLE " << tableNameDb << endl
              << "FIELDS TERMINATED BY '" << fs << "'" << endl
              << "LINES  TERMINATED BY '" << rs << "'" << _cmdTerm << endl <<
              endl;

            tName += "+";
            istat = stat(tName.c_str(), &statbuf);      
        }
}


DbSybase::DbSybase(SchemaMapping& schemaMapping, const string& dbName) :
  Db(schemaMapping, dbName)
{

    _cmdTerm.push_back('\n');
    _cmdTerm.push_back('G');
    _cmdTerm.push_back('O');

    _exec = "$SYBASE/bin/isql";
    _userOption = "-U";
    _passOption = "-P";

    _dbCommand = _exec + " " + _execOption + _userOption +
      "$dbuser" + " " + _passOption + "$dbpw <";

    _envDbUser = "NDB_XDBUSER";
    _envDbPass = "NDB_XDBPW";

}


DbSybase::~DbSybase()
{

}


void DbSybase::WriteLoadingStart(ostream& io)
{

    io << "#!/bin/csh -f" << endl << endl;

    io << "set dbuser = $" << GetEnvDbUser() << endl;
    io << "set dbpw = $" << GetEnvDbPass() << endl;

    io << endl;

}


void DbSybase::WriteLoadingTable(ostream& io, const string& tableName,
  const string& workDir)
{
    string fs;
    string rs;
    escapeString(fs, _fieldSeparator);
    escapeString(rs, _rowSeparator);

        io << "if (-e " << tableName << ".bcp && ! -z " 
	  << tableName << ".bcp) then" << endl;
        io << "   echo \"Loading table " << tableName << "\" "<< endl;
        io << "   $SYBASE/bin/bcp " << _dbName << ".." << tableName <<
          " in " << tableName << ".bcp -c -U$dbuser -P$dbpw -b 10000" <<
          " -t '" << fs  << "'" << " -r '" << rs  << "'" << endl;
        io << "endif" << endl << endl;
}


DbOutput::DbOutput(Db& db) : _db(db)
{

}


DbOutput::~DbOutput()
{

}


void DbOutput::WriteData(Block& block, const string& workDir)
{

}


void DbOutput::WriteSchema(const string& workDir)
{

}


void DbOutput::WriteDataLoadingScripts(const string& workDir)
{

}


void DbOutput::WriteEmptyNumeric(ostream& io)
{

}


void DbOutput::WriteEmptyString(ostream& io)
{

}


void DbOutput::WriteEmptyDate(ostream& io)
{

}


void SqlOutput::WriteEmptyDate(ostream& io)
{

    WriteEmptyString(io);

}


void BcpOutput::WriteEmptyString(ostream& io)
{

    io << _stringDelimiter << _stringDelimiter;

}


void SqlOutput::WriteEmptyString(ostream& io)
{

    io << "NULL";

}


void DbOutput::WriteSpecialChar(ostream& io, const char& specChar)
{

    io << specChar << specChar;

}


void DbOutput::WriteSpecialDateChar(ostream& io, const char& specDateChar)
{


}


void BcpOutput::WriteSpecialDateChar(ostream& io, const char& specDateChar)
{

    if (specDateChar == '"')
        _db.WriteBcpDoubleQuotes(io);

}


void XmlOutput::WriteSpecialChar(ostream& io, const char& specChar)
{

    if (specChar == '>')
        io << "&gt;";
    else if (specChar == '<')
        io << "&lt;";
    else if (specChar == '\'')
        io << "&apos;";
    else if (specChar == '\"')
        io << "&quot;";
    else if (specChar == '&')
        io << "&amp;";
    else if (specChar == '%')
        io << "&#37;"; 

}


void DbOutput::_FormatNumericData(ostream &io,  const string &cs)
{

    if (CifString::IsEmptyValue(cs))
    {
        WriteEmptyNumeric(io); 
        return;
    }

    unsigned int len = cs.size();
    for (unsigned int i = 0; i < len; ++i)
    {
        if (isspace(cs[i]))
        {
            // Skip any white space
            continue;
        }
        else
        {
            io << cs[i];
        }
    }

}


void SqlOutput::WriteEmptyNumeric(ostream& io)
{

    io << "NULL";

}


void DbOutput::_FormatStringData(ostream &io,  const string &cs,
  unsigned int maxWidth)
{

    // A null string or the special CIF characters are treated as NULL . 
    if (CifString::IsEmptyValue(cs))
    {
        WriteEmptyString(io);
        return;
    }

    io << _stringDelimiter;

    unsigned int len = cs.size();

    if (len > maxWidth)
        len = maxWidth;

    bool inLeadWhiteSpace = true;

    for (unsigned int i = 0; i < len; ++i)
    {
        // Skip the leading white space
        if (isspace(cs[i]) && inLeadWhiteSpace)
        {
            continue;
        }
        else if (isspace(cs[i]))
        {
            // Convert all white space to SPACE
            io << " ";
            inLeadWhiteSpace = false;
        }
        else if (IsSpecialChar(cs[i]))
        {
            // Double up double quotes
            WriteSpecialChar(io, cs[i]);
            inLeadWhiteSpace = false;
        }
        else
        {
            io << cs[i];
            inLeadWhiteSpace = false;
        }
    }

    io << _stringDelimiter;

}


void DbOutput::_FormatDateData(ostream &io,  const string &cs,
  unsigned int maxWidth)
{

    if (CifString::IsEmptyValue(cs))
    {
        WriteEmptyDate(io);
        return;
    }

    io << _dateDelimiter;

    unsigned int len = cs.size();

    if (len > maxWidth)
        len = maxWidth;

    bool inLeadWhiteSpace = true;

    for (unsigned int i = 0; i < len; ++i)
    {
        if (isspace(cs[i]) && inLeadWhiteSpace)
        {
            continue;
        }
        else if (isspace(cs[i]))
        {
            io << " ";
            // VLAD - SHOULD WE PUT inLeadWhiteSpace = false???
        }
        else if (IsSpecialDateChar(cs[i]))
        {
            // Double up double quotes
            WriteSpecialDateChar(io, cs[i]);
            inLeadWhiteSpace = false;
        }
        else
        {
            io << cs[i];
            inLeadWhiteSpace = false;
        }
    }

    io << _dateDelimiter;

}


void DbOutput::_FormatTextData(ostream &io,  const string &cs)
{

    // A null string or the special CIF characters are treated as NULL . 
    if (CifString::IsEmptyValue(cs))
    {
        WriteEmptyString(io);
        return;
    }

    io << _stringDelimiter;

    unsigned int len = cs.size();

    for (unsigned int i = 0; i < len; ++i)
    {
        if ((i == 0) && (cs[i] == '\n') && IsFirstTextNewLineSpecial())
	    continue;
        else if (isspace(cs[i]) && cs[i] != '\n')
            // convert all white space to SPACE, except newlines
	    io << " ";
        else if (cs[i] == '\n')
        {
            if ((i != 0) && (cs[i - 1] == '\n'))
                WriteNewLine(io, true);
            else
                WriteNewLine(io);
        }
        else if (IsSpecialChar(cs[i]))
            WriteSpecialChar(io, cs[i]);
        else
	    io << cs[i];
    }

    io << _stringDelimiter;

}


bool DbOutput::IsSpecialChar(const char& character)
{

    vector<char>::const_iterator result = find(_specialChars.begin(),
      _specialChars.end(), character);

    if (result == _specialChars.end())
        return(false);
    else
        return(true);

}


bool DbOutput::IsSpecialDateChar(const char& character)
{

    vector<char>::const_iterator result = find(_specialDateChars.begin(),
      _specialDateChars.end(), character);

    if (result == _specialDateChars.end())
        return(false);
    else
        return(true);

}


void DbOutput::WriteHeader(ostream& io)
{

    io << "#!/bin/csh -f" << endl << endl;

    io << "set dbuser = $" << _db.GetEnvDbUser() << endl;
    io << "set dbpw = $" << _db.GetEnvDbPass() << endl;

}


BcpOutput::BcpOutput(Db& db) : DbOutput(db)
{

    _stringDelimiter = _db.GetBcpStringDelimiter();

    _specialChars.push_back('"');

    _specialDateChars.push_back('"');

    _itemSeparator = _db.GetFieldSeparator();

    _rowSeparator = _db.GetRowSeparator();

}


BcpOutput::~BcpOutput()
{

}


SqlOutput::SqlOutput(Db& db) : DbOutput(db)
{

    _SCHEMA_FILE = "DB_LOADER_SCHEMA.sql";

    _stringDelimiter = "'";

    _specialChars.push_back('\'');

    _dateDelimiter = "'";

    _itemSeparator = ",\n";

}


SqlOutput::~SqlOutput()
{

}


XmlOutput::XmlOutput(Db& db) : DbOutput(db)
{

    _SCHEMA_FILE = "DB_LOADER_SCHEMA_XML.xsd";

    _specialChars.push_back('>');
    _specialChars.push_back('<');
    _specialChars.push_back('\'');
    _specialChars.push_back('"');
    _specialChars.push_back('&');
    _specialChars.push_back('%');

}

XmlOutput::~XmlOutput()
{

}

void DbLoader::Clear() {
   

  _verbose       = false;

  _blockName.clear();

}


void DbLoader::_OpenLog(const string& logName)
{
  if (_verbose && !logName.empty())
      _log.open(logName.c_str(),ios::out|ios::trunc);
}



DbLoader::DbLoader(SchemaMapping& schemaMapping, DbOutput& dbOutput,
  bool verbose, const string& workDir) : _schemaMapping(schemaMapping),
  _dbOutput(dbOutput)
{

  Clear();

  _workDir = workDir;
#ifdef DB_HASH_ID
  _hashMode=1;
#endif
  _verbose = verbose;

  // Set default values  ...  These can be overridden if necessary .

  _blockName = "loadable";

  if (_verbose) {
    string logName = _workDir;
    logName += _LOG_FILE;
    _OpenLog(logName);
  }

}


DbLoader::~DbLoader()
{
    if (_verbose)
        _log.close();
}


void DbLoader::SetWorkDir(const string& workDir)
{

    if (workDir.empty())
        return;

    _workDir = workDir;

}


void Db::SetUseOnlyPopulated(bool mode)
{
    _useOnlyPopulated = mode;
}


bool Db::GetUseOnlyPopulated()
{
    return(_useOnlyPopulated);
}


#ifdef DB_HASH_ID
void DbLoader::SetHashMode(int mode)
{
    _hashMode=mode;
}
#endif


void XmlOutput::_GetTableNameAbbrevXML(string &abbrev, string target)
{
  int len,i,j;
  char name[_MAX_NAME_LENGTH];
  char tmp[_MAX_NAME_LENGTH];

  strcpy(name,target.c_str());
  len=strlen(name);
  memset(tmp,'\0',_MAX_NAME_LENGTH);

  j=0;
  for (i=0; i < len; i++) {
    if (name[i] == '[' || name[i] == ']' || name[i] == '%' || name[i] == '<' || name[i] == '>') 
      continue;
    else { 
      tmp[j]=name[i]; j++;
    }
  }
  tmp[j] = '\0';
  abbrev.clear(); abbrev = tmp;
  return;
}


void XmlOutput::_GetAttributeNameAbbrevXML(string& abbrev,
  const string& targetAttribute)
{

    abbrev.clear();

    unsigned int len = targetAttribute.size();

    for (unsigned int i = 0; i < len; ++i)
    {
        if (targetAttribute[i] == '[' || targetAttribute[i] == ']' ||
          targetAttribute[i] == '%' || targetAttribute[i] == '<' ||
          targetAttribute[i] == '>') 
        {
            continue;
        }
        
        abbrev.push_back(targetAttribute[i]);
    }

}


void DbLoader::AsciiFileToDb(const string& inpFile, const eConvOpt convOpt) 
{

    _INPUT_FILE = inpFile;

    CifFile* fobjR = NULL;

    if (convOpt != eSCRIPTS_ONLY)
    {
        fobjR = new CifFile(_verbose, StringCompare::eCASE_SENSITIVE,
          SchemaMapping::_MAX_LINE_LENGTH);  

        if (_verbose)
            _log << "Reading input file  " << inpFile << endl;

        CifParser* cifParserP = NULL;
        cifParserP = new CifParser(fobjR, fobjR->GetVerbose());

        string diags;
        cifParserP->Parse(inpFile, diags);

        delete(cifParserP);

        if (!diags.empty())
        {
            if (_verbose) _log << " Diagnostics [" << diags.size() << "] " <<
              diags.c_str() << endl;
            diags.clear();
        }
    }

    FileObjToDb(*fobjR, convOpt);

    if (convOpt != eSCRIPTS_ONLY)
    {
        fobjR->Close();
        delete fobjR;
    }

}


void DbLoader::SerFileToDb(const string& inpObjFile, const eConvOpt convOpt)
{

    _INPUT_FILE = inpObjFile;

    CifFile* fobjR = NULL;

    if (convOpt != eSCRIPTS_ONLY)
    {
        fobjR = new CifFile(READ_MODE, inpObjFile, _verbose,
          StringCompare::eCASE_SENSITIVE, SchemaMapping::_MAX_LINE_LENGTH);
    }

    FileObjToDb(*fobjR, convOpt);

    if (convOpt != eSCRIPTS_ONLY)
    {
        fobjR->Close();
        delete fobjR;
    }

}


void DbLoader::FileObjToDb(CifFile& fobjR, const eConvOpt convOpt)
{

    CifFile* fobjW = new CifFile(_verbose, StringCompare::eCASE_SENSITIVE,
      SchemaMapping::_MAX_LINE_LENGTH);  

    fobjW->AddBlock(_blockName);
    _schemaMapping.CreateTables(*fobjW, _blockName);

    if (_verbose)
        _log << "Done with _CreateTable()" << endl;

    if (convOpt != eSCRIPTS_ONLY)
    {
        vector<string> blockNames;
        fobjR.GetBlockNames(blockNames);

        Block& wBlock = fobjW->GetBlock(_blockName);

        unsigned int numBlocks = blockNames.size();
        for (unsigned i = 0; i < numBlocks; ++i)
        {
            if (blockNames[i].empty())
            {
                cerr << "Skipping unnamed block " << endl;
                continue;
            }

            if (_verbose)
                _log << "Loading block " << i + 1 << " " << blockNames[i] <<
                  endl;
 
#ifdef DB_HASH_ID
            if (_hashMode == 1)
            {
                fobjR->GetAttributeValueIf(_HASH_ID, blockNames[i],
                  "database_2", "database_code", "database_id", "PDB");

                if (_HASH_ID.empty())
                {
                    fobjR->GetAttributeValueIf(_HASH_ID, blockNames[i],
                      "database_2", "database_code", "database_id", "NDB");
                }

                if (_HASH_ID.empty())
                {
                    cerr << "Skipping block with no hashable ID " <<
                      blockNames[i] <<endl;
                    continue;
	        }
            }
            else
            {
	        _HASH_ID = blockNames[i];
            }

            if (_verbose)
                _log << "Hash id for block " << i+1 << " " <<
                  blockNames[i] << " is " << _HASH_ID <<endl;
#endif

            Block& rBlock = fobjR.GetBlock(blockNames[i]);

            _LoadBlock(rBlock, wBlock);
        }

        if (_verbose)
            _log << "Done with _LoadBlock()" << endl;

        _dbOutput.WriteData(wBlock, _workDir);
    }

    fobjW->Close();
    delete fobjW;

    if (convOpt != eDATA_ONLY)
    {
        _dbOutput.WriteDataLoadingScripts(_workDir);
    }

}


void DbLoader::_LoadBlock(Block& rBlock, Block& wBlock)
{

    // Create tables in the target schema ...

    // List of tables to be updated in the current schema ... 
    vector<string> tList;

    _schemaMapping.GetTableNames(tList);

    for (unsigned int i = 0; i < tList.size(); ++i)
    {
        if (_verbose)
        {
            _log << " -------------------------------------------"\
              "--------------------" << endl;
            _log << "Loading table "  << i+1 << " " << tList[i] <<
              " of " << tList.size() << endl;
        }

        ISTable* t = wBlock.GetTablePtr(tList[i]);
        if (t == NULL)
        {
            if (_verbose) _log << "Missing output table " << tList[i] << endl;
            continue;
        }

        // For every item (i.e. attribute) of the category (i.e. table), get
        // the info.
        vector<ATTRIBINFO> aI;

        _schemaMapping.GetAttributeInfo(aI, t);

        vector<string> cList;

        // Find all the schema defined attributes of the table and put
        // them in cList
        _schemaMapping.GetAttributeNames(cList, tList[i]);
        if (cList.empty())
        {
            continue;
        }

        unsigned int nCols = cList.size();

        if (_verbose)
            _log << "Schema defines " << nCols << " attributes " << endl;

        vector<vector<string> > mappedAttrInfo;
        _schemaMapping.GetMappedAttributesInfo(mappedAttrInfo, tList[i]);
        if (mappedAttrInfo.empty())
        {
            if (_verbose)
                _log << "No mapped attributes for " << tList[i] << endl;      
            continue; 
        }

        // Get mapped attribute names of this table
        const vector<string>& cNameMap = mappedAttrInfo[0];

        // Get source CIF item names of the mapped attribute names of this table
        const vector<string>& iNameMap = mappedAttrInfo[1];

        // Get condition id and function id of the mapped attribute names of
        //  this table
        const vector<string>& cIdMap = mappedAttrInfo[2];
        const vector<string>& fIdMap = mappedAttrInfo[3];

        // Number of mapped attributes of this table
        unsigned int nColsMap = cNameMap.size();


        //
        // get column indices of mapped attributes in the table.
        //
        vector<int> indMap;
        indMap.insert(indMap.begin(), nColsMap, 0);

        unsigned int ierr = 0;
        for (unsigned int j = 0; j < nColsMap; ++j)
        {
            if (!t->IsColumnPresent(cNameMap[j]))
            {
	        if (_verbose)
                    _log << "Data table missing mapped attribute " <<
                      cNameMap[j] << endl; 
                ierr += 1;
                indMap[j] = -1;
            }
            else
                indMap[j] = DbLoader::GetTableColumnIndex(*t, cNameMap[j]);
        }

        if (ierr)
        {
            continue;
        }

        if (_verbose)
        {
            _log << "Found "  << nColsMap << " attributes in schema map" <<
              endl;
            for (unsigned int j = 0; j < nColsMap; ++j)
            {
	        _log << "Map entry " << j << " is schema attribute " <<
                  indMap[j] << " " << cNameMap[j] <<
                  " mapped to "  << iNameMap[j] << " " <<
                  cIdMap[j] << " " << fIdMap[j] << endl;
            }
        }

        //
        // Temporary space for extracted data isomorphorous with the table t ...
        //   
        //
        vector<vector<string> > dMap;

        vector<string> tmpVector;
        for (unsigned int j = 0; j < nColsMap; ++j)
        {
            dMap.push_back(tmpVector);
        }

        bool iUpdate = false;

        // For every mapped attribute
        for (unsigned int j = 0; j < nColsMap; ++j)
        {
            if (_verbose)
                _log << endl << "*" << endl << "Mapped attribute " << j <<
                  " is " << iNameMap[j] << " condition " <<
                  cIdMap[j] << " function  " << fIdMap[j] <<
                  " schema attribute index " << indMap[j] << endl;

            string tableName;
            CifString::GetCategoryFromCifItem(tableName, iNameMap[j]);

            ISTable* t = rBlock.GetTablePtr(tableName);

            bool updated = _Search(dMap, j, t, rBlock.GetName(), cNameMap,
              iNameMap[j], cIdMap[j], fIdMap[j]);

            if (updated)
                iUpdate = true;

            // We cannot delete the table, since another source attribute
            // for a mapped attribute may be in that table
            // rBlock.DeleteTable(tableName);
        }

        if (!iUpdate || dMap[0].empty() || dMap[1].empty())
        {
            continue;
        }

        // Determine maximum length of non-empty mapped attribute columns
        unsigned int maxLen = dMap[0].size();

        for (unsigned int j = 0; j < nColsMap; ++j)
        {
            if (dMap[j].empty())
                continue;

            if (dMap[j].size() > maxLen)
                maxLen = dMap[j].size();
        }

        // Extend the size of non-empty mapped attribute columns, which are
        // shorter than the maximum size.
        for (unsigned int j = 0; j < nColsMap; ++j)
        {
            if (aI[indMap[j]].iIndex != 1)
                continue;  // hack... if we are not an index.

            if (!dMap[j].empty() && (dMap[j].size() < maxLen))
            {
                if (_verbose)
                    _log << " Extending  map column " << j <<
                      " with " << dMap[j][0]  << endl;

                for (unsigned int k = dMap[j].size(); k < maxLen; ++k)
                { 
                    dMap[j].push_back(dMap[j][0]);
                }
            }
        }


        // Check consistency. Minimum and maximum length must be the same.

        unsigned int minLen = dMap[0].size();
        maxLen = dMap[0].size();

        for (unsigned int j = 0; j < nColsMap; ++j)
        {
            if (dMap[j].empty())
                continue;

            if (dMap[j].size() < minLen)
                minLen = dMap[j].size();

            if (dMap[j].size() > maxLen)
                maxLen = dMap[j].size();
        }

        if (minLen != maxLen)
        {
            if (_verbose)
            {
                _log << "Conflict in length min = " << minLen <<
                  " max = " << maxLen << endl;
                _log << "Skipping table update update" << endl;
            }

            cerr << "In " << _INPUT_FILE << ": " <<
              "Skipping update for table " << tList[i] <<
              " with inconsistent column lengths." << endl;

            continue;
        }
      
        //
        //  Mapping OK, Update table ... 
        //
        if (_verbose)
            _log << " Table length = " << minLen << endl;

#ifndef DEBUG_POINT
    if (t->GetName() == "citation")
    {
        int a = 1;
        int b = a + 3;
        ++b;
    }
#endif

        // Here minLen and maxLen are the same and represent the number
        // of rows.
        for (unsigned int j = 0; j < minLen; ++j)
        {
            bool iskip = false;

            for (unsigned int k = 0; k < nColsMap; ++k)
            {
                if ((dMap[k].empty() || dMap[k][j].empty()) &&
                  (aI[indMap[k]].iIndex || aI[indMap[k]].iNull))
                {
                    iskip = true;

                    if (_verbose)
                       _log << "Skipping row with NULL value in key "\
                         "attribute column " << indMap[k] << endl;

                    cerr  << "In " << _INPUT_FILE << ": "  <<
                      "Skipping row in " << tList[i] <<
                      " with NULL value in key column " << indMap[k] <<
                      endl;

                    break;
                }
            }

            if (iskip)
                continue; 
	    
            if (_verbose)
            {
                unsigned int iRow = t->GetNumRows();
                _log << "Updating " << tList[i] << " row " <<
                  iRow << endl;
            }

            vector<string> row(nCols, String::Empty);

	    for (unsigned int k = 0; k < nColsMap; ++k)
            {
	        if (dMap[k].empty())
                    continue;

	        if (_verbose)
                    _log << "Updating attribute cell " << indMap[k] <<
                      " value " << dMap[k][j] << endl;

                // Cell must not be used if it is located in "ndb_id" column
                // and its value starts with "RCSB", because that indicates
                // a condition where RCSB id is stored in "ndb_id" column and
                // is to be ignored. In all other situations the cell must be
                // used.
                if ((cNameMap[k] != "ndb_id") || (dMap[k][j].compare(0, 4,
                  "RCSB", 0, 4) != 0))
                {
                    // Use the cell.
	            row[indMap[k]] = dMap[k][j];
                }
	    }

	    t->AddRow(row);
        }

        wBlock.WriteTable(t);
    }

}


void DbLoader::_DoFunc(vector<string>& s, const vector<string>& r,
  const string& sFnct)
{

    // VLAD - This method changes (if appropriate) every element of vector "r"
    // according to the function sFnct and stores each element in vector "s"

    s.clear();

    string cs;
    char *p = NULL;
    int i,iOK;
    time_t curr_time;
    struct tm *tmptr = NULL;

    if (sFnct.empty() || (sFnct == CifString::NullValue) ||
      (sFnct == "__missing()"))
    {
        // Empty string, unknown or internal missing attribute handler.
        if (r.empty())
        {
            s.push_back("");
        }
        else
        {
            s = r;
        }
    }
    else if (sFnct == "today()")
    {
        p = new char[18];
        time(&curr_time);
        tmptr = localtime(&curr_time);
        strcpy(p,"");
        strftime(p,18,"%Y-%b-%d %R", tmptr);
        s.push_back(p);
        if (p)
            delete [] p;
    }
    else if (sFnct == "prefix(PDB)")
    {
        if (r.empty())
        {
            s.push_back("");
        }
        else
        {
            for (int i=0; i < (int) r.size(); i++)
            {
                cs = "PDB" + r[i];
                s.push_back(cs);
            }
        }
    }
    else if (sFnct == "prefixUp(PDB)")
    {
        if (r.empty())
        {
            s.push_back("");
        }
        else
        {
            for (int i=0; i < (int) r.size(); i++)
            {
                cs = "PDB" + r[i];
                _ToUpperString(cs);
                s.push_back(cs);
            }
        }
    }
    else if (sFnct == "date()")
    {
        if (r.empty())
        {
            s.push_back("");
        }
        else
        {
            string dbDate;
            for (unsigned int i = 0; i < r.size(); ++i)
            {
	        _dbOutput._db.ConvertDate(dbDate, r[i]);
	        s.push_back(dbDate);
            }
        }
    }
    else if (sFnct == "timestamp()")
    {
        if (r.empty())
        {
            s.push_back("");
        }
        else
        {
            string dbTimestamp;
            for (unsigned int i = 0; i < r.size(); ++i)
            {
	        _dbOutput._db.ConvertTimestamp(dbTimestamp, r[i]);
	        s.push_back(dbTimestamp);
            }
        }
    }
    else if (sFnct == "switchname(comma)")
    {
        if (r.empty())
        {
            s.push_back("");
        }
        else
        {
            int ilen = 0;
            for (unsigned int i = 0; i < r.size(); i++)
            {
                ilen += r[i].size() + 3;
            }

            char *buf = new char[ilen + 1];
            memset(buf, 0, ilen + 1);

            char* p = new char[ilen + 1];
            memset(p, 0, ilen + 1);

            for (unsigned int i = 0; i < r.size(); i++)
            {
                strcpy(p, r[i].c_str());
                name_conversion_first_last(p);
                strcat(buf, p);
                if (i < r.size() - 1)
                    strcat(buf, ", ");
            }
            string cs = buf;
            delete [] buf;
            delete [] p;
            s.push_back(cs);
        }
    }
    else if (sFnct == "tointegerplus()")
    {
        if (r.empty())
        {
            s.push_back("");
        }
        else
        {
            for (unsigned int i = 0; i < r.size(); i++)
            {
                if (String::IsCiEqual(r[i], "primary"))
                {
                    s.push_back(String::IntToString(1));
                }
                else if (String::IsNumber(r[i]))
                {
                    int number = String::StringToInt(r[i]);
                    s.push_back(String::IntToString(number + 1));
                }
                else
                {
                    // Default to empty string
                    s.push_back("");
                }
            }
        }
    }
#ifdef VLAD_DATE_OBSOLETE
    else if (sFnct == "datecnv(sybase)")
    {
        if (r.empty())
        {
            s.push_back("");
        }
        else
        {
            char sybdate[20];
            for (int i=0; i < (int) r.size(); i++)
            {
	        _dformat_3(r[i].c_str(), sybdate, 0);
	        s.push_back(sybdate);
            }
        }
    }
    else if (sFnct == "datecnv(sybase-short)")
    {
        if (r.empty())
        {
            s.push_back("");
        }
        else
        {
            char sybdate[20];
            for (int i=0; i < (int) r.size(); i++)
            {
	        _dformat_3(r[i].c_str(),sybdate, 1);
	        s.push_back(sybdate);
            }
        }
    }
    else if (sFnct == "datecnv(oracle-short)")
    {
        if (r.empty())
        {
            s.push_back("");
        }
        else
        {
            char sybdate[20];
            for (int i=0; i < (int) r.size(); i++)
            {
	        _dformat_5(r[i].c_str(),sybdate);
	        s.push_back(sybdate);
            }
        }
    }
#endif // VLAD_DATE_OBSOLETE
    else if (sFnct == "row()")
    {
        if (r.empty())
        {
            s.push_back("");
        }
        else
        {
            for (int i=1; i <= (int) r.size(); i++)
            {
	        cs = String::IntToString(i);
	        s.push_back(cs);
            }
        }
    }
    else if (sFnct == "toupper()")
    {
        if (r.empty())
        {
            s.push_back("");
        }
        else
        {
            for (int i=0; i < (int) r.size(); i++)
            {
	        cs = r[i];
	        _ToUpperString(cs);
	        s.push_back(cs);
            }
        }
    }
    else if ((sFnct == "strip(ws)") || (sFnct == "strip()"))
    {
        if (r.empty())
        {
            s.push_back("");
        }
        else
        {
            for (int i=0; i < (int) r.size(); i++)
            {
	        cs = r[i];
	        _StripString(cs,1);
	        s.push_back(cs);
            }
        }
    }
    else if (sFnct == "strip(nl)")
    {
        if (r.empty())
        {
            s.push_back("");
        }
        else
        {
            for (int i=0; i < (int) r.size(); i++)
            {
	        cs = r[i];
               _StripString(cs,2);
	       s.push_back(cs);
            }
        }
    }
    else if (sFnct == "collapse(comma)")
    {
        if (r.empty())
        {
            s.push_back("");
        }
        else
        {
            int ilen  = 0;
            for (i=0; i < (int) r.size(); i++)
            {
	        ilen += r[i].size() + 3; 
            }
            char *buf = (char *) calloc(ilen+1, sizeof(char));
            strcpy(buf,"");
            for (i=0; i < (int) r.size(); i++)
            {
	        strcat(buf, r[i].c_str());
	        if (i < (int) r.size()-1)
                    strcat(buf,", ");
            }
            cs.clear(); cs = buf;
            free(buf);
      
            s.push_back(cs);
        }
    }
    else if (sFnct == "collapse(comma,cnvname)")
    {
        if (r.empty())
        {
            s.push_back("");
        }
        else
        {
            int ilen  = 0;
            for (i=0; i < (int) r.size(); i++)
            {
	        ilen += r[i].size() + 2; 
            }
            char *buf = (char *) calloc(ilen+1, sizeof(char));
            strcpy(buf,"");
            for (i=0; i < (int) r.size(); i++)
            {
                string reorderedName;
        	_ReorderName(reorderedName, (char*)(r[i].c_str()),1);
        	strcat(buf, reorderedName.c_str());
        	if (i < (int) r.size()-1) strcat(buf,", ");
            }
            cs.clear(); cs = buf;
            free(buf);
      
            s.push_back(cs);
        }
    }
    else if (sFnct == "collapse(space)")
    {
        if (r.empty())
        {
            s.push_back("");
        }
        else
        {
            int ilen  = 0;
            for (i=0; i < (int) r.size(); i++)
            {
	        ilen += r[i].size() + 2; 
            }
            char *buf = (char *) calloc(ilen+1, sizeof(char));
           strcpy(buf,"");
           for (i=0; i < (int) r.size(); i++)
           {
	       strcat(buf, r[i].c_str());
	       strcat(buf," ");
           }
           cs.clear(); cs = buf;
           free(buf);      

           s.push_back(cs);
        }
    }
    else if (sFnct == "collapse()")
    {
        if (r.empty())
        {
            s.push_back("");
        }
        else
        {
            int ilen  = 0;
            for (i=0; i < (int) r.size(); i++)
            {
	        ilen += r[i].size(); 
            }
            char *buf = (char *) calloc(ilen+1, sizeof(char));
            strcpy(buf,"");
            for (i=0; i < (int) r.size(); i++)
            {
	        strcat(buf, r[i].c_str());
            }
            cs.clear(); cs = buf;
            free(buf);      

            s.push_back(cs);
        }
    }
    else if (sFnct == "count()")
    {
        if (r.empty())
        {
            s.push_back("0");
        }
        else
        {
            cs = String::IntToString((int)r.size());
            s.push_back(cs);
        }
    }
    else if (sFnct == "ifany()")
    {
        // one or more non-null values ...
        if (r.empty())
        {
            s.push_back("N");
        }
        else
        {
            iOK=0;
            for (i=0; i < (int) r.size(); i++)
            {
	        if (!CifString::IsEmptyValue(r[i]))
                    iOK++;
            }
            if (iOK) 
	        s.push_back("Y");
            else
                s.push_back("N");
        }
    }
    else if (sFnct == "tobool()")
    {
        // convert y/yes -> 1  anyother,NULL,missing -> 0
        if (r.empty())
        {
            s.push_back("0");
        }
        else
        {
            for (i=0; i < (int) r.size(); i++)
            {
	        if (!CifString::IsEmptyValue(r[i]))
                {
	            if (String::IsCiEqual(r[i], "Y") ||
                      String::IsCiEqual(r[i], "yes")) 
	                s.push_back("1");
	            else
	                s.push_back("0");
	        }
                else
                {
                    //  NULLS map to 0
                    s.push_back("0");
	        }
            }
        }
    }
    else if (sFnct.compare(0, 9, "ifexists(", 9) == 0)
    {
        // one or more non-null values ...
        if (r.empty())
        {
            s.push_back("N");
        }
        else
        {
            iOK=0;
            p = new char[strlen(sFnct.c_str()) + 1];
            strcpy(p,&sFnct.c_str()[9]);
            p[strlen(p)-1] = '\0';
            cs.clear(); cs = p;

            for (i=0; i < (int) r.size(); i++)
            {
	        if (!CifString::IsEmptyValue(r[i]))
                {
                    iOK++;
	            s.push_back(cs);
	        }
                else
                {
	            s.push_back("");	  
	        }
            }
            if (p) delete p;
        }
    }
    else if (sFnct == "first()")
    {
        // first() of possibly multiple values
        if (r.empty())
        {
            s.push_back("");
        }
        else
        {
            s.push_back(r[0]);
        }
    }
    else if (sFnct == "last()")
    {
        // last() of possibly multiple values
        if (r.empty())
        {
            s.push_back("");
        }
        else
        {
	  s.push_back(r[r.size()-1]);
        }
    }

    else
    {
        if (r.empty())
        {
            s.push_back("");
        }
        else
        {
            s = r;
        }
    }

}


int DbLoader::_GetMapColumnIndex(const vector<string>& cNameMap,
  const string& vOf) 
{
  char *p=NULL;
  string cs1;
  int j, jCol=-1;

  if ( !strncmp(vOf.c_str(),"valueof(",8) || !strncmp(vOf.c_str(),"unseqof(",8) ) {
    p = new char[strlen(vOf.c_str()) + 1];
    strcpy(p,&(vOf.c_str()[8]));
    p[strlen(p)-1] = '\0';
    cs1.clear(); cs1 = p;
    if (p) delete p;
    if (_verbose) _log << "Map join attribute is " << cs1 << endl;
    jCol = -1;
    for (j=0; j < (int) cNameMap.size(); j++) {
      if (String::IsCiEqual(cs1, cNameMap[j])) {
	jCol = j; 
	break;
      }
    }
    if (jCol < 0) {
      if (_verbose) _log << "Attribute " << cs1 << " is not in map" << endl;
    }
  }
  return(jCol);
}


void DbLoader::_GetMapColumnValue(string& p,
  vector<vector<string> >& dMap,  int iCol,  int irow)
{
  p.clear();

  int nRows = 0, len = 0;

  if (_verbose) _log << " ** Fetching values from map column " << iCol << " row " << irow << endl;

  if (iCol < 0 ) return; 
  if (irow < 0 || (irow >= (int) dMap[iCol].size())) return;
  nRows= dMap[iCol].size();  

  if (_verbose) _log << " ** Fetching value from map column " << iCol << " row " << irow 
		     << " of " << nRows << endl;

  len = dMap[iCol][irow].size();
  if (len > 0) {
    p = dMap[iCol][irow];
  } 

}


int DbLoader::_GetMapColumnLength(vector<vector<string> >& dMap,   int iCol)
{
  int len=0;
  if (iCol < 0 ) return(0);
  if (dMap[iCol].empty()) return(0);
  len = dMap[iCol].size(); 
  if (_verbose) {
    _log << " ** Length of map column " << iCol << " is " << len << endl; 
  }
  return(len);
}


bool DbLoader::_Search(
  vector<vector<string> >& dMap,  // container for output
  const unsigned int iAttrib,     // index between map and schema
  ISTable* isTableP,              // source table
  const string& blockName,        // block name 
  const vector<string>& cNameMap, // mapped attributes 
  const string& sItem,            // source item
  const string& sCnd,
  const string& sFnct             // condition and function code
)
{

    if (blockName.empty())
        return(false);

    //
    // jdw catch the special cases of a constant non-schema mapping
    //

    vector<string> r;

    if (sFnct == "today()")
    {
        if (_verbose)
            _log << "Constant function today()" << endl;

        _DoFunc(dMap[iAttrib], r, sFnct);
        if (_verbose)
        {
            _log << "Returning result length " << dMap[iAttrib].size() << endl;
            for (unsigned int i = 0; i < dMap[iAttrib].size(); ++i)
            {
	        _log << "Returning value " << dMap[iAttrib][i] << endl;
            }
        }
        return(true);
    }
    else if (sFnct == "datablockid()")
    {
        if (_verbose)
            _log << "Constant function datablockid()" << endl;
        vector<string> s;
        s.push_back(blockName);
        dMap[iAttrib] = s;

        if (_verbose)
        {
            _log << "Returning result length " << dMap[iAttrib].size() << endl;
            for (unsigned int i = 0; i < dMap[iAttrib].size(); ++i)
            {
	        _log << "Returning value " << dMap[iAttrib][i] << endl;
            }
        }
        return(true);
    }
#ifdef DB_HASH_ID
    else if (sFnct == "unseq()")
    {
        // unseq() unique sequence using pdb hash code

        long long hashId;
        hashId = pdbIdHash(_HASH_ID);

        if (_verbose)
        {
            _log << "Function seqid()" << endl;
            _log << "Block id "<< blockName << endl;
            _log << "Hash id "<< (long)hashId  << endl;
        }

        vector<string> s;
        unsingned int maxLen = 0;
        for (i=0; i < iAttrib; i++)
        {
            if (dMap[i].empty())
                continue;
            j=dMap[i].size();
            if (j > maxLen)
                maxLen = j;
        }

        if (maxLen>0)
        {
            for (i=1; i <= maxLen; i++)
            {
	        csTarget.clear();
                csTarget += String::IntToString((long long) hashId +
                  (long long) i);
	        if (_verbose)
                    _log << "hashid csTarget " << csTarget << endl;
                s.push_back(csTarget);
            }
        }
        else
        {
            s.push_back("");
        }
        dMap[iAttrib]=s;
        return(true);
    }
#endif
    else if (sFnct == "row()")
    {
        if (_verbose) _log << "Function row()" << endl;
        vector<string> s;
        unsigned int maxLen = 0;
        for (unsigned int i = 0; i < (unsigned int) iAttrib; ++i)
        {
            if (dMap[i].empty())
                continue;
            unsigned int j = dMap[i].size();
            if (j > maxLen)
                maxLen = j;
        }

        if (maxLen>0)
        {
            for (unsigned int i = 1; i <= maxLen; i++)
            {
                string csTarget;
                csTarget += String::IntToString((int) i);
                s.push_back(csTarget);
            }
        }
        else
        {
            s.push_back("");
        }
        dMap[iAttrib]=s;
        return(true);
    }
#ifdef VLAD_HASH_ID_DEL
    else
    {
        return(false);
    }
#endif

    string columnName;
    string tableName;
    CifString::GetItemFromCifItem(columnName, sItem);
    CifString::GetCategoryFromCifItem(tableName, sItem);
    if (sItem.empty() || columnName.empty() || tableName.empty())
        return(false);


    if (_verbose)
        _log << "Searching block " << blockName << " table " <<
          tableName << " column " << columnName << endl;

    if (isTableP == NULL)
    {
        if (_verbose)
            _log << "Target table " << tableName << " not in " <<
              blockName << endl;

        return(false);
    }

#ifndef DEBUG_POINT
    if (isTableP->GetName() == "citation_author")
    {
        int a = 1;
        int b = a + 3;
        ++b;
    }
#endif

    if (!isTableP->IsColumnPresent(columnName))
    {
        if (_verbose)
            _log << "Target columnName " << columnName << " not in " <<
              tableName << endl; 
        vector<string> s;
        for (unsigned int i = 0; i < isTableP->GetNumRows(); ++i)
        {
            s.push_back(CifString::NullValue);
        }
        dMap[iAttrib] = s;
        if (_verbose)
        {
            _log << "Returning result length " << dMap[iAttrib].size() << endl;
            for (unsigned int i = 0; i < dMap[iAttrib].size(); ++i)
            {
	        _log << "Returning value " << dMap[iAttrib][i] << endl;
            }
        }
        return(true);
    }

    if (!sCnd.empty())
    {
        // fetch condition 

        if (_verbose)
            _log << "Using search condition " << sCnd << endl;

        vector<vector<string> > mappedConditions; 
        _schemaMapping.GetMappedConditions(mappedConditions, sCnd);

        if (!mappedConditions.empty())
        {
            // list of constraints for condition.

            vector<string> cndCol;
            cndCol = mappedConditions[0];

            vector<string> cndVal;
            cndVal = mappedConditions[1];

            unsigned int nEqCnd=0;
            for (unsigned int i = 0; i < mappedConditions[0].size(); ++i)
            {
                if ( !strcmp( cndVal[i].c_str(),"datablockid()"))
                {
	            cndVal[i].clear(); cndVal[i] = blockName;
	        }

	        if ( !strncmp( cndVal[i].c_str(),"valueof(",8) ||
                  !strncmp( cndVal[i].c_str(),"unseqof(",8) )
                {
	            nEqCnd++;
	        }

	        if (_verbose)
                    _log << "Condition column " << cndCol[i] <<
                      " value " << cndVal[i] << endl;
            }

            if (nEqCnd)
            {
                // number of join conditions.

                int* indDMap = NULL;
                int* lenDMap = NULL;
                indDMap = new int[mappedConditions[0].size()];
                lenDMap = new int[mappedConditions[0].size()];
                int lenMin = -1, lenMax = -1;

                for (unsigned int i = 0; i < mappedConditions[0].size(); ++i)
                {
	            indDMap[i] = -1;
	            lenDMap[i] =  0;
	            if ( !strcmp( cndVal[i].c_str(),"datablockid()"))
                    {
	                cndVal[i].clear(); cndVal[i] = blockName;
	            }
	            if ( !strncmp( cndVal[i].c_str(),"valueof(",8) ||
                      !strncmp( cndVal[i].c_str(),"unseqof(",8))
                    {
	                indDMap[i] = _GetMapColumnIndex(cNameMap, cndVal[i]);
	                lenDMap[i] = _GetMapColumnLength(dMap,indDMap[i]);
	                if (lenMin < 0 || lenDMap[i] < lenMin)
                            lenMin  = lenDMap[i];
	                if (lenDMap[i] > lenMax)
                            lenMax  = lenDMap[i];
	                if (_verbose)
                            _log << " ** value or unseqof() index " <<
                              indDMap[i] << " length " << lenDMap[i] << endl;
	            }
	        }
	        if (lenMin != lenMax)
                {
	            if (_verbose)
                        _log << " ** valueof() or unseqof() column length "\
                          "inconsistency  lenMin " <<  lenMin <<
                          " lenMax " << lenMax << endl;
	        }
	        if (_verbose)
                    _log << " ** valueof() or unseqof() column lenMin " <<
                      lenMin << " lenMax " << lenMax << endl;

                if (!cndCol.empty())
                    cndCol.clear();
	        if (!cndVal.empty())
                    cndVal.clear();

                vector<string> tRes;

	        for (unsigned int j = 0; (int)j < lenMin; ++j)
                {
                    cndCol = mappedConditions[0];
                    cndVal = mappedConditions[1];
                    if (_verbose)
                        _log << " ** Starting row "<< j <<
                          " condition length " << mappedConditions[0].size() << 
                          endl;

                    unsigned int iFlagValueOf = 0;
	            for (unsigned int i = 0; i < mappedConditions[0].size();
                      ++i)
                    {
	                if (_verbose) _log << " ** condition value " << i <<
                          " is " << cndVal[i] << endl;
	    
	                if ( !strcmp( cndVal[i].c_str(),"datablockid()"))
                        {
	                    cndVal[i].clear(); cndVal[i] = blockName;
	                }
	    
	                if ( !strncmp( cndVal[i].c_str(),"valueof(",8) ||
                          !strncmp( cndVal[i].c_str(),"unseqof(",8) )
                        {
                            iFlagValueOf = 1;
	                    if (!strncmp( cndVal[i].c_str(),"unseqof(",8))
                                iFlagValueOf = 2; 
                            string p;
	                    _GetMapColumnValue(p, dMap, indDMap[i], j);
	                    if (p.empty())
                            {
		                cndVal[i].clear(); 
		                cndVal[i] = "NULL";
		                if (_verbose)
                                    _log << " ** Map row " <<  j <<
                                      " has value NULL" << endl;
	                    }
                            else
                            {
		                cndVal[i].clear(); 
		                cndVal[i] = p;
		                if (_verbose)
                                    _log << " ** Map row " <<  j <<
                                      " has value " << p << endl;
	                    }
                        } 
	                if (_verbose)
                            _log << " ** Search column " <<
                              cndCol[i] << " for " <<
                              cndVal[i] << endl;
	            }

                    vector<unsigned int> is;
	            isTableP->Search(is, cndVal, cndCol);

	            if ( !is.empty())
                    {
	                if (iFlagValueOf == 1)
                        {
                            // valueof()
	                    if (_verbose)
                                _log << " ** Search result length is " <<
                                  is.size() << " rows"<< endl;
	                    isTableP->GetColumn(r, columnName,is);
	                }
#ifdef DB_HASH_ID
                        else if (iFlagValueOf == 2)
                        {
                            // unseqof()
	                    if (_verbose)
                                _log << " ** Search result length is " <<
                                  is.size() << " rows"<< endl;
                            r.clear();
                            long long hashId;
	                    hashId = pdbIdHash(_HASH_ID);
	                    for (int jj = 0; jj < (int) is.size(); jj++)
                            {
		                cs.clear();
                                cs += String::IntToString((long long) (is[jj] +
                                  1 + hashId));
                                r.push_back(cs);
                            }
	                }
#endif
	            }
                    else
                    {
	                if (_verbose)
                        {
	                    _log << " ** ** ** ** ** Warning " << endl;
	                   if (is.empty())
                               _log << " ** Search returns 0 length" << endl;
	                }
	                r.clear();
	            }

                    vector<string> r1;
	            _DoFunc(r1, r, sFnct);
	            tRes.push_back(r1[0]);
	            r1.clear();

	            if (!cndVal.empty()) cndVal.clear();
	            if (!cndCol.empty()) cndCol.clear();
	            if (!r.empty()) r.clear();
	        } // end j loop	
                // copy expand the resulting column
	        _DoFunc(dMap[iAttrib], tRes, CifString::NullValue);

	        if (indDMap) delete indDMap;
	        if (lenDMap) delete lenDMap;
            }
            else
            {
                // If no join conditions nEqCnd ...
                vector<unsigned int> is;
                isTableP->Search(is, cndVal, cndCol);
                if (!is.empty())
                {
                    isTableP->GetColumn(r, columnName, is);
                }
	        _DoFunc(dMap[iAttrib], r, sFnct);
            }
        }
        else
        {  // search condition missing in category
            isTableP->GetColumn(r, columnName);
            _DoFunc(dMap[iAttrib], r, sFnct);
        }
    }
    else
    { // no condition ...
        if (_verbose)
            _log << "No search condition specified, selecting column " <<
                columnName << endl;
        isTableP->GetColumn(r, columnName);
        if (_verbose && r.empty())
        {
            _log << "Column "<< columnName << " returns NULL result." << endl;
        }
        else if (_verbose && !r.empty())
        {
            _log << "Column "<< columnName << " returns length " <<
              r.size() << endl;
        }

        _DoFunc(dMap[iAttrib], r, sFnct);
    }

    if (_verbose)
    {
        if (!dMap[iAttrib].empty())
        {
            _log << "Returning result length " << dMap[iAttrib].size() << endl;
            for (unsigned int i = 0; i < dMap[iAttrib].size(); ++i)
            {
	        _log << "Returning " << columnName << " value " <<
                  dMap[iAttrib][i] << endl;
            }
        }
        else
        {
            _log << "Returning result length 0 " << endl;
        }
    }

    r.clear();

    return (true);

}


void SchemaMapping::CreateTables(CifFile& fobj, const string& blockName)
{

    // Create tables in the target schema ...

    vector<string> tNames;
    GetTableNames(tNames);

    for (unsigned int i = 0; i < tNames.size(); ++i)
    {
        vector<string> cNames;
        GetAttributeNames(cNames, tNames[i]);

        if (cNames.empty())
        {
            continue;
        }

        ISTable* t = new ISTable(tNames[i]);
        for (unsigned int j = 0; j < cNames.size(); ++j)
        {
            t->AddColumn(cNames[j]);
        }

        Block& block = fobj.GetBlock(blockName);
        block.WriteTable(t);
    }

}


void SqlOutput::WriteAuxTables(ostream& io, ISTable* infoP,
  const vector<string>& tableNames)
{

    string auxTableNameDb;
    _db._schemaMapping.GetTableNameAbbrev(auxTableNameDb, infoP->GetName());

    for (unsigned int i = 0; i < tableNames.size(); ++i)
    {
        string tableNameDb;

        _db._schemaMapping.GetTableNameAbbrev(tableNameDb, tableNames[i]);

        _db.WriteDeleteTable(io, auxTableNameDb, "tablename", tableNameDb);

        io << endl;
    }

    _WriteTable(io, infoP, true);

}


void SqlOutput::WriteSchema(const string& workDir)
{

    string oFile = workDir;
    oFile += _SCHEMA_LOADING_SCRIPT;
    ofstream ioc(oFile.c_str(), ios::out | ios::trunc);

    WriteSqlScriptSchemaInfo(ioc);

    ioc.close();

    oFile = workDir;
    oFile += _SCHEMA_FILE;
    ofstream io(oFile.c_str(),ios::out|ios::trunc);

    _db.WriteSchemaStart(io);
    io << endl << endl;

    oFile = workDir;
    oFile += _SCHEMA_DELETE_FILE;
    ofstream iod(oFile.c_str(), ios::out | ios::trunc);

    string start;
    _db.GetStart(start);
    iod << start << endl << endl;

    //
    // Generate SQL to create table schema...
    //
    vector<string> tableNames;
    _db._schemaMapping.GetTableNames(tableNames);
    for (unsigned int i = 0; i < tableNames.size(); ++i)
    {
        if (tableNames[i].empty())
            continue;

        string tableNameDb;
        _db._schemaMapping.GetTableNameAbbrev(tableNameDb, tableNames[i]);

        vector<AttrInfo> attrInfo;
        _db._schemaMapping.GetAttributesInfo(attrInfo, tableNames[i]);
        if (attrInfo.empty())
        {
            continue;
        } 

        if (_db.GetUseOnlyPopulated() &&
          (!_db._schemaMapping.IsTablePopulated(attrInfo)))
        {
            continue;
        } 

        _db.DropTableSql(iod, tableNameDb);
        iod << endl;

        CreateTableSql(io, tableNames[i], attrInfo);
        io << endl << endl;
    }

    iod.close();

    //
    // Create tableinfo and columninfo tables which contain data about schema.
    //
    ISTable* tinfo = _db._schemaMapping.CreateTableInfo();

    WriteAuxTables(io, tinfo, tableNames);

    io << endl;

    if (tinfo != NULL)
        delete tinfo;

    ISTable* cinfo = _db._schemaMapping.CreateColumnInfo();

    WriteAuxTables(io, cinfo, tableNames);

    if (cinfo != NULL)
        delete cinfo;

    io.close();

}


void XmlOutput::WriteSchema(const string& workDir)
{

    //
    // File for XML XSD output ...
    //
    string oFile = workDir;
    oFile += _SCHEMA_FILE;
    ofstream io(oFile.c_str(), ios::out | ios::trunc);

    io << "<?xml version=\"1.0\"?>" << endl;
    io << "<xsd:schema xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\" >" <<
      endl;

#ifdef VLAD_PREVIOUSLY_COMMENTED_CODE
    //     << "            targetNamespace = \"http://deposit.pdb.org/mmcif/pdbx\"" << endl
    //     << "            xmlns = \"http://deposit.pdb.org/mmcif/pdbx\" >" << endl;
#endif

    //
    //
    // Generate XML complex types...
    //

    vector<string> tableNames;
    _db._schemaMapping.GetTableNames(tableNames);

    for (unsigned int i = 0; i < tableNames.size(); ++i)
    {
        _CreateComplexTypeXML(io, tableNames[i]);
    }
 
    io << "      <xsd:complexType name=\"datablockType\">" << endl;
    io << "        <xsd:all>" << endl;

    string  tableNameDb;
    for (unsigned int i = 0; i < tableNames.size(); ++i)
    {
        _GetTableNameAbbrevXML(tableNameDb, tableNames[i]);
        if (!strncmp("pdbx_",tableNameDb.c_str(), 5))
        {
            io <<  "          <xsd:element ref=\"category_" << tableNameDb <<
              "\"" << " minOccurs=\"0\" maxOccurs=\"1\" />" << endl;
        }
        else if (!strncmp("ndb_",tableNameDb.c_str(), 4))
        {
            io <<  "          <xsd:element ref=\"category_" << tableNameDb <<
              "\"" <<  " minOccurs=\"0\" maxOccurs=\"1\" />" << endl;
        }
        else
        {
            io <<  "          <xsd:element ref=\"category_" << tableNameDb <<
              "\"" <<  " minOccurs=\"0\" maxOccurs=\"1\" />" << endl;
        }
    }

    io << "        </xsd:all>" << endl;  
    io << "        <xsd:attribute name=\"datablockName\" " ;
    io << "type=\"xsd:string\" " ;
    io << "use=\"optional\" ";
    io << " />" << endl;
    io << "      </xsd:complexType>" << endl;


    for (unsigned int i = 0; i < tableNames.size(); ++i)
    {
        _GetTableNameAbbrevXML(tableNameDb, tableNames[i]);

        if (!strncmp("pdbx_",tableNameDb.c_str(), 5))
        {
            io <<  "      <xsd:element name=\"category_" << tableNameDb <<
              "\"" << "  type=\"" << tableNameDb << "Type\" />" << endl;
        }
        else if (!strncmp("ndb_",tableNameDb.c_str(), 4))
        {
            io <<  "      <xsd:element name=\"category_" << tableNameDb <<
              "\"" << "  type=\"" << tableNameDb << "Type\" />" << endl;
        }
        else
        {
            io <<  "      <xsd:element name=\"category_" << tableNameDb <<
              "\"" << "  type=\"" << tableNameDb << "Type\" />" << endl;
        }
    }

    io << "      <xsd:element name=\"datablock\" " <<
      "  type=\"datablockType\" />" << endl;

    io << "</xsd:schema>" << endl;

    io.close();

}


void XmlOutput::_CreateComplexTypeXML(ostream &io, const string& tableName)
{

    int i;
    string columnName, dType, tableNameDb, columnNameDb;
    int iIndex;

    if (tableName.size() <= 0) return;
    if (!io )   return;
    _GetTableNameAbbrevXML(tableNameDb, tableName);

    io << endl;

    if (!strncmp("pdbx_", tableNameDb.c_str(), 5))
    {
        io <<  "      <xsd:complexType name=\"" << tableNameDb  <<
          "Type\">" << endl;
        _SetNamespaceDefaultXML("pdbx"); 
    }
    else if (!strncmp("ndb_", tableNameDb.c_str(), 4))
    {
        io <<  "      <xsd:complexType name=\"" << tableNameDb <<
          "Type\">" << endl;
        _SetNamespaceDefaultXML("ndb"); 
    }
    else
    {
        io <<  "      <xsd:complexType name=\"" << tableNameDb <<
          "Type\">" << endl;
        _SetNamespaceDefaultXML("mmcif"); 
    }

    //  Add new stuff 
    io << "        <xsd:sequence>" << endl;
    io << "        <xsd:element name=\"" << tableNameDb << "\"" 
     <<  " minOccurs=\"0\" maxOccurs=\"unbounded\" >" <<   endl;
    io << "        <xsd:complexType>" << endl;


    // Get attribute level information
    //
    vector<AttrInfo> attrInfo;
  
    _db._schemaMapping.GetAttributesInfo(attrInfo, tableName);
    if (attrInfo.empty())
        return;

    io << "        <xsd:sequence>" << endl;

    for (i=0; i < (int) attrInfo.size(); i++)
    {
        columnName = attrInfo[i].attribName;
        _GetAttributeNameAbbrevXML(columnNameDb, columnName);

        dType = attrInfo[i].dataType;
        iIndex     = atoi((attrInfo[i].indexFlag).c_str());
        // iNull      = atoi((attrInfo[i].nullFlag).c_str());
        // iWidth     = atoi((attrInfo[i].width).c_str());
        //      iPrecision = atoi((attrInfo[i].precision).c_str());

        if (iIndex)
        {
            // index items are displayed as attributes.
	    continue;
        }

        io << "            <xsd:element name=\"" ;
        _QualifyNameXML(io,columnNameDb); 
        io << "\" ";
        io << "type=\"" ;
        _ConvertDataTypeXML(io,dType);
        io << "\" ";
        io << " minOccurs=\"0\" maxOccurs=\"1\" nillable=\"true\" ";
        io << " />" << endl;
    }

    io <<  "        </xsd:sequence>"   << endl;

    //  indices as attributes

    for (i=0; i < (int) attrInfo.size(); i++)
    {
        columnName = attrInfo[i].attribName;
        _db._schemaMapping.GetAttributeNameAbbrev(columnNameDb, tableName,
          columnName);

        dType = attrInfo[i].dataType;
        iIndex     = atoi((attrInfo[i].indexFlag).c_str());
        // iNull      = atoi((attrInfo[i].nullFlag).c_str());
        // iWidth     = atoi((attrInfo[i].width).c_str());
        // iPrecision = atoi((attrInfo[i].precision).c_str());

        if (!iIndex)
        {
	    continue;
        }

        io << "        <xsd:attribute name=\"" ;

        _QualifyNameXML(io,columnNameDb);

        io << "\" ";
        io << "type=\"" ;

        _ConvertDataTypeXML(io,dType);

        io << "\" ";
        io << "use=\"required\"";
        io << " />" << endl;
    }

    io <<  "      </xsd:complexType>" << endl;
    io <<  "      </xsd:element>" << endl;
    io <<  "      </xsd:sequence>" << endl;

    io <<  "      </xsd:complexType>" << endl << endl;
    //  END of TYPE DEFINITION

}


static void _ConvertDataTypeXML(ostream &io,  string dType)
{

  if (String::IsCiEqual(dType, "char") || String::IsCiEqual(dType, "text")) {
    dType.clear();
    dType = "xsd:string";
  } 

  if (String::IsCiEqual(dType, "float") || String::IsCiEqual(dType, "double") ) {
    dType.clear();
    dType = "xsd:decimal";
  } 

  if (String::IsCiEqual(dType, "int") || String::IsCiEqual(dType, "integer") || String::IsCiEqual(dType, "bigint")) {
    dType.clear();
    dType = "xsd:integer";
  } 

  if (String::IsCiEqual(dType, "date") || String::IsCiEqual(dType, "datetime")) {
    dType.clear();
    dType = "xsd:date";
  } 

  io << dType;

}


void XmlOutput::_QualifyNameXML(ostream& io, const string& name)
{

    for (unsigned int i = 0; i < name.size(); ++i)
    {
        if ((name[i] == '[') || (name[i] == ']') || (name[i] == '%') ||
          (name[i] == '<') || (name[i] == '>'))
        {
            continue;
        }

        io << name[i];
    }

}


void SchemaMapping::GetAllAttributeInfo(vector<ATTRIBINFO>& aI,
  string tableName) 
{
  // Return array of ATTRIBINFO for the name input table  .
  aI.clear();

  int  nattrib;
  int i, j, nCols;
  string columnName;

  vector<AttrInfo> attrInfo;
  
  GetAttributesInfo(attrInfo, tableName);
  if (attrInfo.empty())
  {
#ifdef VLAD_LOG_SEPARATION
    if (_verbose) _log << "GetAllAttributeInfo: No attributes for found for " << tableName << endl;
#endif
      return;
  }

  nCols = attrInfo.size();
#ifdef VLAD_LOG_SEPARATION
  if (_verbose) _log << "GetAllAttributeInfo: +++ TABLE " << tableName 
		     << " columns " << nCols << endl;
#endif
  if (!nCols)  return;

  ATTRIBINFO tmpAttrInfo;
  aI.insert(aI.begin(), nCols, tmpAttrInfo);
  for (j=0; j < nCols; j++) {
    aI[j].dType.clear(); 
    aI[j].iIndex     = 0;
    aI[j].iNull      = -1;
    aI[j].iWidth     = -1;
    aI[j].iPrecision = -1;
    aI[j].iTypeCode  = -1;
    aI[j].iMaxWidth  = 0;
  }

  //
  // Get all of the attributes for this table.
  //

  nattrib=0;
  for (i=0; i < (int) attrInfo.size(); i++) {
    {
      columnName = attrInfo[i].attribName;
      j = i;
#ifdef VLAD_LOG_SEPARATION
      if (_verbose) _log << "GetAllAttributeInfo: Column [" << j << "] " <<  columnName << endl;
#endif
      if (j >= 0  && j < nCols) {
        aI[j].dType = attrInfo[i].dataType;
	if ( aI[j].dType.c_str() !=NULL &&  String::IsCiEqual(aI[j].dType, "char")) {
	  aI[j].iTypeCode = SchemaMapping::_TYPE_CODE_STRING;
	} else 	if ( aI[j].dType.c_str()!=NULL &&  String::IsCiEqual(aI[j].dType, "int")) {
	  aI[j].iTypeCode = SchemaMapping::_TYPE_CODE_INT;
	} else 	if ( aI[j].dType.c_str()!=NULL &&  String::IsCiEqual(aI[j].dType, "float")) {
	  aI[j].iTypeCode = SchemaMapping::_TYPE_CODE_FLOAT;
	} else 	if ( aI[j].dType.c_str()!=NULL && String::IsCiEqual(aI[j].dType, "text")) {
	  aI[j].iTypeCode = SchemaMapping::_TYPE_CODE_TEXT;
	} else 	if ( aI[j].dType.c_str()!=NULL &&  String::IsCiEqual(aI[j].dType, "datetime")) {
	  aI[j].iTypeCode = SchemaMapping::_TYPE_CODE_DATETIME;
	} else 	if ( aI[j].dType.c_str()!=NULL &&  String::IsCiEqual(aI[j].dType, "bigint")) {
	  aI[j].iTypeCode = SchemaMapping::_TYPE_CODE_BIGINT;
	}
	
	aI[j].iIndex     = atoi((attrInfo[i].indexFlag).c_str());
	aI[j].iNull      = atoi((attrInfo[i].nullFlag).c_str());
	aI[j].iWidth     = atoi((attrInfo[i].width).c_str());
	aI[j].iPrecision = atoi((attrInfo[i].precision).c_str());


#ifdef VLAD_LOG_SEPARATION
	if (_verbose) _log << " Width " << aI[j].iWidth << " Type  " << aI[j].iTypeCode 
                           << " Index " << aI[j].iIndex << " Null " << aI[j].iNull << endl;
#endif
        nattrib++;
      } 
    }

  }
  if (nattrib < nCols) {
#ifdef VLAD_LOG_SEPARATION
    if (_verbose) _log << "GetAllAttributeInfo(): warning unmapped column nattrib = " << nattrib  << endl;
#endif
  }
#ifdef VLAD_LOG_SEPARATION
  if (_verbose) _log << "GetAllAttributeInfo: +++ DONE WITH " << tableName << endl << endl;
#endif

}



void SchemaMapping::GetAttributeInfo(vector<ATTRIBINFO>& aI, ISTable* tIn)
{
  // Return array of ATTRIBINFO for the input table.
  aI.clear();

  string tableNameC;
  int  nattrib;
  int i, j, nCols;
  string columnName, tableName;

  if (!tIn ) return;

  tableNameC = tIn->GetName();
  if (tableNameC.empty()) return;

  nCols = tIn->GetNumColumns();
  if (!nCols)  return;
  ATTRIBINFO tmpAttrInfo;
  aI.insert(aI.begin(), nCols, tmpAttrInfo);
  for (j=0; j < nCols; j++) {
    aI[j].dType.clear(); 
    aI[j].iIndex     = 0;
    aI[j].iNull      = -1;
    aI[j].iWidth     = -1;
    aI[j].iPrecision = -1;
    aI[j].iTypeCode  = -1;
    aI[j].iMaxWidth  = 0;
  }

  //
  // Get all of the attributes for this table.
  //


  tableName = tableNameC;

  vector<AttrInfo> attrInfo;
  
  GetAttributesInfo(attrInfo, tableName);
  if (attrInfo.empty())
  {
#ifdef VLAD_LOG_SEPARATION
    if (_verbose) _log << "GetAttributeInfo: No attributes for found for " << tableNameC << endl;
#endif
      return;
  }

#ifdef VLAD_LOG_SEPARATION
  if (_verbose) _log << "GetAttributeInfo: +++ TABLE " << tableNameC << endl << endl;
#endif

  nattrib=0;
  for (i=0; i < (int) attrInfo.size(); i++) {
    {
      columnName = attrInfo[i].attribName;

      j = DbLoader::GetTableColumnIndex(*tIn, columnName);      
#ifdef VLAD_LOG_SEPARATION
      if (_verbose) _log << "GetAttributeInfo: Column [" << j << "] " <<  columnName << endl;
#endif
      if (tIn->IsColumnPresent(columnName)) {
        j = DbLoader::GetTableColumnIndex(*tIn, columnName);
        aI[j].dType = attrInfo[i].dataType;
	if ( aI[j].dType.c_str()!=NULL &&  String::IsCiEqual(aI[j].dType, "char")) {
	  aI[j].iTypeCode = SchemaMapping::_TYPE_CODE_STRING;
	} else 	if ( aI[j].dType.c_str()!=NULL &&  String::IsCiEqual(aI[j].dType, "int")) {
	  aI[j].iTypeCode = SchemaMapping::_TYPE_CODE_INT;
	} else 	if ( aI[j].dType.c_str()!=NULL && String::IsCiEqual(aI[j].dType, "float")) {
	  aI[j].iTypeCode = SchemaMapping::_TYPE_CODE_FLOAT;
	} else 	if ( aI[j].dType.c_str()!=NULL && String::IsCiEqual(aI[j].dType, "text")) {
	  aI[j].iTypeCode = SchemaMapping::_TYPE_CODE_TEXT;
	} else 	if ( aI[j].dType.c_str()!=NULL && String::IsCiEqual(aI[j].dType, "datetime")) {
	  aI[j].iTypeCode = SchemaMapping::_TYPE_CODE_DATETIME;
	} else 	if ( aI[j].dType.c_str()!=NULL && String::IsCiEqual(aI[j].dType, "bigint")) {
	  aI[j].iTypeCode = SchemaMapping::_TYPE_CODE_BIGINT;
	}
	
	aI[j].iIndex     = atoi((attrInfo[i].indexFlag).c_str());
	aI[j].iNull      = atoi((attrInfo[i].nullFlag).c_str());
	aI[j].iWidth     = atoi((attrInfo[i].width).c_str());
	aI[j].iPrecision = atoi((attrInfo[i].precision).c_str());


#ifdef VLAD_LOG_SEPARATION
	if (_verbose) _log << " Width " << aI[j].iWidth << " Type  " << aI[j].iTypeCode 
                           << " Index " << aI[j].iIndex << " Null " << aI[j].iNull << endl;
#endif
        nattrib++;
      } 
    }

  }
  if (nattrib < nCols) {
#ifdef VLAD_LOG_SEPARATION
    if (_verbose) _log << "GetAttributeInfo(): warning unmapped column nattrib = " << nattrib  << endl;
#endif
  }
#ifdef VLAD_LOG_SEPARATION
  if (_verbose) _log << "GetAttributeInfo: +++ DONE WITH " << tableNameC << endl << endl;
#endif


}


bool SchemaMapping::AreAttributesMandatory(const vector<string>& row,
  const vector<ATTRIBINFO>& aI)
{

    if (aI.empty())
        return(false);

    for (unsigned int i = 0; i < row.size(); ++i)
    {
        if ((aI[i].iNull == 0) && (aI[i].iIndex == 0))
            continue;

        if (CifString::IsEmptyValue(row[i]))
            return(false);
    }

    return(true);

}


bool SchemaMapping::IsTablePopulated(const vector<AttrInfo>& attrInfo)
{

    for (unsigned int i = 0; i < attrInfo.size(); ++i)
    {
        if (attrInfo[i].populated == "Y")
            return (true);
    }

    return (false);

}


int DbLoader::_CheckNullRow(const vector<string>& row,
  const vector<ATTRIBINFO>& aI)
{

    if (aI.empty())
        return(0);

    for (unsigned int i = 0; i < row.size(); ++i)
    {
        if ((aI[i].iNull == 1) || (aI[i].iIndex == 1))
            continue;

        if (CifString::IsEmptyValue(row[i]))
            continue;

        return(0);
    }

    return(1);
}


int SchemaMapping::CheckIndexAttributes(const vector<string>& row,
  const vector<ATTRIBINFO>& aI)
{

    if (row.empty() || aI.empty())
        return(0);

    for (unsigned int i = 0; i < row.size(); ++i)
    {
        if (aI[i].iIndex == 0)
            continue;

        if (CifString::IsEmptyValue(row[i]))
            return(0);
    }
    return(1);
}


void DbOutput::SetInputFile(const string& inpFile)
{
    _INPUT_FILE = inpFile;
}


void XmlOutput::WriteData(Block& block, const string& workDir)
{

    cerr << "Starting cif->xml" << endl;

    string oFile = workDir;
    if (!_INPUT_FILE.empty())
    {
        oFile += _INPUT_FILE;
        oFile += ".xml";
    }
    else
    {
        oFile += "data.xml";
    }
 
    cerr << "CIF->XML for file " << oFile << endl;

    ofstream ioxml;
    if (_db.GetAppendFlag())
    {
        ioxml.open(oFile.c_str(), ios::out | ios::app);
    }
    else
    {
        ioxml.open(oFile.c_str(), ios::out | ios::trunc);
    }

    ioxml << "<?xml version=\"1.0\"?>" << endl;
    ioxml <<  "<datablock  xmlns=\"http://deposit.pdb.org/mmcif/pdbx\"" <<
      endl << "             xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" >" << endl;

    vector<string> tableNames;
    _db._schemaMapping.GetTableNames(tableNames);

    for (unsigned int i = 0; i < tableNames.size(); ++i)
    {

        string tableNameDb;
        _db._schemaMapping.GetTableNameAbbrev(tableNameDb, tableNames[i]);

        vector<AttrInfo> attrInfo;
        _db._schemaMapping.GetAttributesInfo(attrInfo, tableNames[i]);
        if (attrInfo.empty())
        {
            continue;
        } 

        if (_db.GetUseOnlyPopulated() &&
          (!_db._schemaMapping.IsTablePopulated(attrInfo)))
        {
            continue;
        } 

        string tableNameXml;
        _GetTableNameAbbrevXML(tableNameXml, tableNames[i]);

        cerr << " Table " << tableNames[i] << " abbreviation " <<
          tableNameXml << endl;

        ioxml <<  "      <category_"  << tableNameXml << ">" << endl;

        if (!strncmp("pdbx_", tableNames[i].c_str(), 5))
        {
            _SetNamespaceDefaultXML("pdbx"); 
        }
        else if (!strncmp("ndb_", tableNames[i].c_str(), 4))
        {
            _SetNamespaceDefaultXML("ndb"); 
        }
        else
        {
            _SetNamespaceDefaultXML("mmcif"); 
        }

        ISTable* t = block.GetTablePtr(tableNames[i]);
        if (t != NULL && t->GetNumRows() > 0)
        {
            _WriteTable(ioxml, t);
        }

        ioxml <<  "      </category_" << tableNameXml << ">" << endl;
    }

    ioxml <<  "</datablock>" << endl;

    ioxml.close();

}


void XmlOutput::_WriteTable(ostream& io, ISTable* tIn, bool noRevise)
{

    // Write table using the rows of tIn  - XML version.

    if (!io || !tIn)
        return;

    // Get all of the attributes for this table.
    vector<ATTRIBINFO> aI;
    _db._schemaMapping.GetAttributeInfo(aI, tIn);

    const string& tableName = tIn->GetName();

    string tableNameDb;
    _GetTableNameAbbrevXML(tableNameDb, tableName);

    unsigned int nRows = tIn->GetNumRows();
    unsigned int nCols = tIn->GetNumColumns();

    const vector<string>& columnNames = tIn->GetColumnNames();

    for (unsigned int i = 0; i < nRows; ++i)
    {
        const vector<string>& row = tIn->GetRow(i);

#ifdef VLAD_LOG_SEPARATION
        if (_verbose)
        {
            _log << "XML writing row " << i << endl;
        }
#endif
        if (!SchemaMapping::AreAttributesMandatory(row, aI))
        {
#ifdef VLAD_LOG_SEPARATION
            if (_verbose)
                _log << "Failing check on row " << i << endl;
#endif
            continue;
        }

        if (row.empty())
        {
            continue;
        }

        io << "         <";
        _QualifyNameXML(io,  tableNameDb);

        for (unsigned int j = 0; j < nCols; ++j)
        {
            // VLAD - THIS HAS TO BE REVISED
            // Define which types can be indices. It seems strange that a
            // text type can be attribute, because of the performance.

            // Write indices as attibutes...
            if (aI[j].iIndex == 0)
            {
                continue;
            }

            string columnNameDb;
            _GetAttributeNameAbbrevXML(columnNameDb, columnNames[j]);

            io << " ";
            _QualifyNameXML(io, columnNameDb);
            io << "=\"";

            _FormatData(io, row[j], aI[j], noRevise);

            io << "\"";
        }
        io << ">"  << endl;

        // elements here ... check if null beforehand...

        for (unsigned int j = 0; j < nCols; ++j)
        {
            if (aI[j].iIndex == 1)
            {
                continue;
            }

            string columnNameDb;
            _GetAttributeNameAbbrevXML(columnNameDb, columnNames[j]);

            io << "            <";
            _QualifyNameXML(io, columnNameDb);

            if (CifString::IsEmptyValue(row[j]))
                io << " xsi:nil=\"true\" ";

            io << ">";

            _FormatData(io, row[j], aI[j], noRevise);

            io << "</";
            _QualifyNameXML(io,columnNameDb);
            io << ">" << endl;
        }
        io << "         </";
        _QualifyNameXML(io, tableNameDb);
        io << ">" << endl;
    }

    if (noRevise)
    {
        return;
    }

#ifdef VLAD_LOG_SEPARATION
    if (_verbose)
    {
        for (i=0; i < nCols; i++)
        {
            _log  << tIn->GetName() << " " << columnNames[i] << " " <<
              aI[i].iMaxWidth << endl;
        }
    }
#endif

    if (_db._schemaMapping.GetReviseSchemaMode())
    {
        for (unsigned int i = 0; i < nCols; ++i)
        {
            _db._schemaMapping.UpdateAttributeDef(tIn->GetName(),
              columnNames[i], aI[i].iTypeCode, aI[i].iWidth, aI[i].iMaxWidth);
        }
    }

}


void XmlOutput::_GetDefaultNamespaceXML(ostream& io) 
{

    io << _NS_DEFAULT << ":";

}


void XmlOutput::_SetNamespaceDefaultXML(const string& name) 
{

    if (!name.empty())
    {
        _NS_DEFAULT = name;
    }
    else
    {
        _NS_DEFAULT = "mmcif";
    }

}


void SchemaMapping::GetMasterIndexAttribName(string& masterIndexAttribName)
{

    // Master index attribute name must be in the first row of
    // "rcsb_attribute_def" table, in "attribute_name" column

    masterIndexAttribName = (*_attribSchema)(0, "attribute_name");

}

void DbOutput::GetMasterIndexAttribValue(string& masterIndexAttribValue,
  Block& block, const string& masterIndexAttribName,
  const vector<string>& tableNames)
{

    for (unsigned int i = 0; i < tableNames.size(); ++i)
    {
        if (tableNames[i].empty())
        {
            continue;
        }

        ISTable* t = block.GetTablePtr(tableNames[i]);
        if ((t != NULL) && (t->GetNumRows() > 0))
        {
            // const vector<string>& row = tIn->GetRow(0);

            // dbIndexValue = row[0];

            masterIndexAttribValue = (*t)(0, masterIndexAttribName);
            break;
        }
    }
}

void BcpOutput::WriteData(Block& block, const string& workDir)
{

    // BCP Update for ...

    string oFile = workDir;
    oFile += _DATA_DELETE_FILE;

    ofstream iosql;
    if (_db.GetAppendFlag())
    {
        iosql.open(oFile.c_str(), ios::out | ios::app);
    }
    else
    {
        iosql.open(oFile.c_str(), ios::out | ios::trunc);
    }

    string start;
    _db.GetStart(start);

    iosql << start << endl << endl;

    vector<string> tableNames;
    _db._schemaMapping.GetTableNames(tableNames);

    string masterIndexAttribName;
    _db._schemaMapping.GetMasterIndexAttribName(masterIndexAttribName);

    string masterIndexAttribValue;
    GetMasterIndexAttribValue(masterIndexAttribValue, block,
      masterIndexAttribName, tableNames);

    for (unsigned int i = 0; i < tableNames.size(); ++i)
    {
        if (tableNames[i].empty() || (tableNames[i] == "rcsb_tableinfo") ||
          (tableNames[i] == "rcsb_columninfo"))
        {
            continue;
        }

        string tableNameDb;
        _db._schemaMapping.GetTableNameAbbrev(tableNameDb, tableNames[i]);

        vector<AttrInfo> attrInfo;
        _db._schemaMapping.GetAttributesInfo(attrInfo, tableNames[i]);
        if (attrInfo.empty())
        {
            continue;
        }

        if (_db.GetUseOnlyPopulated() &&
          (!_db._schemaMapping.IsTablePopulated(attrInfo)))
        {
            continue;
        }

        _db.WriteDeleteTable(iosql, tableNameDb, masterIndexAttribName,
          masterIndexAttribValue);

        iosql << endl;

        ISTable* t = block.GetTablePtr(tableNames[i]);
        if ((t != NULL) && (t->GetNumRows() > 0))
        {
            string tName = workDir;
            tName += tableNames[i];
            tName += ".bcp";

            struct stat statbuf;
            int istat = stat(tName.c_str(), &statbuf);

            while (istat == 0 && (statbuf.st_size > _MAXFILESIZE))
            {
                tName += "+";
                istat = stat(tName.c_str(), &statbuf);
                cerr << "File size for " << tName << " is " << statbuf.st_size 
                  << " istat " << istat << endl;
            }

            ofstream iobcp;
            if (_db.GetAppendFlag())
            {
                iobcp.open(tName.c_str(), ios::out | ios::app);
            }
            else
            {
                iobcp.open(tName.c_str(), ios::out | ios::trunc);
            } 

            _WriteTable(iobcp, t);

            iobcp.close();
        }
    }

    iosql.close();

}


void BcpOutput::WriteDataLoadingScripts(const string& workDir)
{

    WriteDataLoadingScript(workDir);

    WriteDataLoadingFile(workDir);

}


void SqlOutput::WriteDataLoadingScripts(const string& workDir)
{

    WriteDataLoadingScript(workDir);

}


void SqlOutput::WriteDataLoadingScript(const string& workDir)
{

    string oFile = workDir;
    oFile += _DATA_LOADING_SCRIPT;

    ofstream io(oFile.c_str(), ios::out | ios::trunc);

    WriteHeader(io);
    io << endl;

    WriteDbExec(io, _DATA_FILE);

    io << endl;

    io.close();
 
}


void SqlOutput::WriteData(Block& block, const string& workDir)
{

    string oFile = workDir;
    oFile += _DATA_FILE;

    ofstream iosql;
    if (_db.GetAppendFlag())
    {
        iosql.open(oFile.c_str(), ios::out | ios::app);
    }
    else
    {
        iosql.open(oFile.c_str(), ios::out | ios::trunc);
    }

    string start;
    _db.GetStart(start);

    iosql << start << endl << endl;

    vector<string> tableNames;
    _db._schemaMapping.GetTableNames(tableNames);

    string masterIndexAttribName;
    _db._schemaMapping.GetMasterIndexAttribName(masterIndexAttribName);

    string masterIndexAttribValue;
    GetMasterIndexAttribValue(masterIndexAttribValue, block,
      masterIndexAttribName, tableNames);

    for (unsigned int i = 0; i < tableNames.size(); ++i)
    {
        if (tableNames[i].empty() || (tableNames[i] == "rcsb_tableinfo") ||
          (tableNames[i] == "rcsb_columninfo"))
        {
            continue;
        }

        string tableNameDb;
        _db._schemaMapping.GetTableNameAbbrev(tableNameDb, tableNames[i]);

        vector<AttrInfo> attrInfo;
        _db._schemaMapping.GetAttributesInfo(attrInfo, tableNames[i]);
        if (attrInfo.empty())
        {
            continue;
        } 

        if (_db.GetUseOnlyPopulated() &&
          (!_db._schemaMapping.IsTablePopulated(attrInfo)))
        {
            continue;
        }

        _db.WriteDeleteTable(iosql, tableNameDb, masterIndexAttribName,
          masterIndexAttribValue);
        iosql << endl;

        ISTable* t = block.GetTablePtr(tableNames[i]);
        if ((t != NULL) && (t->GetNumRows() > 0))
        {
            _WriteTable(iosql, t);
        }
        iosql << endl;
    }

    iosql.close();

}


void Db::WriteLoad(ostream& io)
{

    io << "source " << GetDataLoadingFileName() << endl;

}


void DbOutput::_FormatStringDataSql(ostream& io, const string& cs,
  unsigned int maxWidth) 
{

    // A null string or the special CIF characters are treated as NULL. 
    if (CifString::IsEmptyValue(cs))
    {
        io << "NULL";
        return;
    }

    unsigned int len = cs.size();

    if (len > maxWidth)
        len = maxWidth;

    io << "'";

    bool inLeadWhiteSpace = true;

    for (unsigned int i = 0; i < len; ++i)
    {
        if (isspace(cs[i]) && inLeadWhiteSpace)
        {
            // Skip leading white space
            continue;
        }
        else if (isspace(cs[i]))
        {
            // Convert all white space to SPACE
            io << " ";
        }
        else if (cs[i] == '\'')
        {
            // Double-up single quotes
            io << "''";
            inLeadWhiteSpace = false;
        }
        else
        {
            io << cs[i];
            inLeadWhiteSpace = false;
        }
    }

    io << "'";

}


void BcpOutput::WriteDataLoadingScript(const string& workDir)
{

    string oFile = workDir;
    oFile += _DATA_LOADING_SCRIPT;

    ofstream io(oFile.c_str(), ios::out | ios::trunc);

    WriteHeader(io);
    io << endl;

    WriteDelete(io);
    io << endl;

    _db.WriteLoad(io);
    io << endl;

    io.close();

}


void BcpOutput::WriteDataLoadingFile(const string& workDir)
{

    ofstream ioctl;
    ioctl.open(_db.GetDataLoadingFileName().c_str(), ios::out | ios::trunc);

    _db.WriteLoadingStart(ioctl);
    ioctl << endl << endl;

    vector<string> tableNames;
    _db._schemaMapping.GetTableNames(tableNames);

    for (unsigned int i = 0; i < tableNames.size(); ++i)
    {
        _db.WriteLoadingTable(ioctl, tableNames[i], workDir);
    }

    _db.WriteLoadingEnd(ioctl);

    ioctl.close();

}


void BcpOutput::WriteDelete(ostream& io)
{

    io << "if (\"x$1\" == \"xdelete\" && -e " << _DATA_DELETE_FILE <<
       " && ! -z " << _DATA_DELETE_FILE << ") then" << endl;

    WriteDbExecOnly(io, _DATA_DELETE_FILE);

    io << "endif" << endl;

}


ISTable* SchemaMapping::CreateTableInfo() 
{

    ISTable* tinfo = NULL;

    const vector<string>& columnNames = _tableSchema->GetColumnNames();

    vector<string> tableNames;
    _tableSchema->GetColumn(tableNames, columnNames[0]);
    vector<string> c1;
    _tableSchema->GetColumn(c1, columnNames[1]);
    vector<string> c2;
    _tableSchema->GetColumn(c2, columnNames[2]);
    vector<string> c3;
    _tableSchema->GetColumn(c3, columnNames[3]);
    vector<string> c4;
    _tableSchema->GetColumn(c4, columnNames[4]);

    Block& rsBlock = _fobjS->GetBlock("rcsb_schema");
    ISTable* td = rsBlock.GetTablePtr("rcsb_table_description");

    tinfo = new ISTable("rcsb_tableinfo");

    tinfo->AddColumn("tablename");
    tinfo->AddColumn("description");
    tinfo->AddColumn("type");
    tinfo->AddColumn("table_serial_no");
    tinfo->AddColumn("group_name");
    tinfo->AddColumn("WWW_Selection_Criteria");  
    tinfo->AddColumn("WWW_Report_Criteria");  

    string cs1 = "ADDB01";
  

    for (unsigned int i = 0; i < tableNames.size(); ++i)
    {
        if ((tableNames[i] == "tableinfo") || (tableNames[i] == "columninfo"))
            continue;
    
#ifdef VLAD_LOG_SEPARATION
        if (_verbose) _log << endl << "_CreateTableInfo  table = " <<
          tableNames[i] << endl;
#endif
    
        string cs3;
        vector<unsigned int> ir;

        td->Search(ir, tableNames[i], columnNames[0]);
        if (!ir.empty())
        {
            const vector<string>& row = td->GetRow(ir[0]);
            if (!row.empty())
            {
	        cs3 = row[1];
            }
        }
        else
        {
#ifdef VLAD_LOG_SEPARATION
            if (_verbose)
                _log << endl << "_CreateTableInfo  NO information for " <<
                  tableNames[i] << endl;
#endif
            // Add default text 
            cs3 = "Missing table description";
        }

        vector<string> newRow;

        string cs2 = String::IntToString(i + 1);

        string tableNameDb;
        GetTableNameAbbrev(tableNameDb, tableNames[i]);
        newRow.push_back(tableNameDb);
        newRow.push_back(cs3);
        newRow.push_back(c1[i]);
        newRow.push_back(cs2);
        newRow.push_back(c2[i]);
        newRow.push_back(c3[i]);
        newRow.push_back(c4[i]);

        tinfo->AddRow(newRow);
    }

    return(tinfo);

}


ISTable* SchemaMapping::CreateColumnInfo() 
{
  string tableNameDb, columnNameDb;

  ISTable *cinfo = NULL;

  const vector<string>& columnNames = _attribSchema->GetColumnNames();

  vector<string> c0;
  _attribSchema->GetColumn(c0, columnNames[0]);
  vector<string> c1;
  _attribSchema->GetColumn(c1, columnNames[1]);
  //  vector<string> *c2 = _attribSchema->GetColumn(2);
  //  vector<string> *c3 = _attribSchema->GetColumn(3);
  //  vector<string> *c4 = _attribSchema->GetColumn(4);
  //  vector<string> *c5 = _attribSchema->GetColumn(5);
  //  vector<string> *c6 = _attribSchema->GetColumn(6);

  Block& rsBlock = _fobjS->GetBlock("rcsb_schema");
  ISTable *td = rsBlock.GetTablePtr("rcsb_attribute_description");

  ISTable *tv = rsBlock.GetTablePtr("rcsb_attribute_view");

  cinfo = new ISTable("rcsb_columninfo");

  cinfo->AddColumn("tablename");
  cinfo->AddColumn("columnname");
  cinfo->AddColumn("description");
  cinfo->AddColumn("example");
  cinfo->AddColumn("type");
  cinfo->AddColumn("table_serial_no");
  cinfo->AddColumn("column_serial_no");
  cinfo->AddColumn("WWW_Selection_Criteria");  
  cinfo->AddColumn("WWW_Report_Criteria");  

  vector<unsigned int> ir;
  vector<string> * q0 = NULL;
  q0 = new   vector<string>;
  vector<string> iq0;
  
  int it=0;
  string cs1, cs2, cs3, cs4, cs5, cs6,cs7, cs8;
  cs1.clear(); 
  cs1 = "ADDB01";
  
  const vector<string>& tdColNames = td->GetColumnNames();
  const vector<string>& tvColNames = tv->GetColumnNames();

  vector<string> newRow;

  for (int i=0; i < (int) c0.size(); i++) {
    if (!strcmp(c0[i].c_str(),"tableinfo") || 
	!strcmp(c0[i].c_str(),"columninfo")) continue;
    if (i > 0 && strcmp(c0[i].c_str(),c0[i-1].c_str())) it++;
    cs8.clear(); cs8 += String::IntToString(it+1);
    cs2.clear(); cs2 += String::IntToString(i+1);
    cs3.clear(); cs4.clear();  cs5.clear(); cs6.clear(); cs7.clear();

    q0->clear(); q0->push_back(c0[i]); q0->push_back(c1[i]);
    iq0.clear(); iq0.push_back(tdColNames[0]);     iq0.push_back(tdColNames[1]);
    td->Search(ir, *q0, iq0);
    if (!ir.empty()) {
      const vector<string>& row = td->GetRow(ir[0]);
      if (!row.empty()) {
	cs3 = row[2];
	cs4 = row[3];
      }    
      ir.clear();
    } 
    if (cs3.size() == 0 ) cs3 = "Missing column description";
    if (cs4.size() == 0 ) cs4 = "Missing column example";

    q0->clear(); q0->push_back(c0[i]); q0->push_back(c1[i]);
    iq0.clear(); iq0.push_back(tvColNames[0]);     iq0.push_back(tvColNames[1]);
    tv->Search(ir, *q0, iq0);
    if (!ir.empty()) {
      const vector<string>& row = tv->GetRow(ir[0]);
      if (!row.empty()) {
	cs5 = row[2];
	cs6 = row[3];
	cs7 = row[4];
      }      
      ir.clear();
    } else {
      cs5 = "3";
      cs6 = "1";
      cs7 = "1";
    }



    newRow.clear();

    GetTableNameAbbrev(tableNameDb, c0[i]);
    newRow.push_back(tableNameDb);

    GetAttributeNameAbbrev(columnNameDb, c0[i], c1[i]);
    newRow.push_back(columnNameDb);

    newRow.push_back(cs3); // description
    newRow.push_back(cs4); // example
    newRow.push_back(cs5);  // type

    newRow.push_back(cs8);  // column serial
    newRow.push_back(cs2);  // column serial
    newRow.push_back(cs6);  // www select
    newRow.push_back(cs7);  // www report

    cinfo->AddRow(newRow); 

  }

  //  if (_verbose) cinfo->PrintTable();
  if (q0 != NULL)
      delete q0;

  return(cinfo);
}


#ifdef VLAD_DATE_OBSOLETE
void DbLoader::_dformat_1(const char *date, char *odate) {
  // reformat  dd-mmm-yy  to Mon dd, yyyy

  if (odate !=NULL) strcpy (odate,"");
  if (date == NULL ||  date[2] != '-' || date[6] != '-' || strlen(date) != 9) return;



  strncpy(odate,&date[3],3);
  odate[1] = tolower(odate[1]);
  odate[2] = tolower(odate[2]);

  strcat(odate," ");
  strncat(odate,&date[0],2);
  if (odate[4] == '0') odate[4] = ' ';

  strcat(odate," ");
  if (date[7] == '0' || date[7] == '1' || date[7] == '2' || date[7] == '3' || date[7] == '4' || date[7] == '5') {
    strcat(odate,"20");
    strncat(odate,&date[7],2);
  } else {
    strcat(odate,"19");
    strncat(odate,&date[7],2);
  }
}


void DbLoader::_dformat_2(const char *date, char *odate) {
  //
  // reformat  CIF data = yyyy-mm-dd[:00:00]  to Mon dd, yyyy
  //                      2000-01-27  to Jan 27, 2000  odate[13]
  //

  if (odate != NULL) strcpy (odate,"");
  if (date == NULL ||  date[4] != '-' || date[7] != '-' ) return;

  char month[5];
  memset(month, 0 ,5);
  strncpy(month,&date[5],2);




  if (!strcmp(month,"01"))
    strcpy(odate,"Jan ");
  else   if (!strcmp(month,"02"))
    strcpy(odate,"Feb ");
  else   if (!strcmp(month,"03"))
    strcpy(odate,"Mar ");
  else   if (!strcmp(month,"04"))
    strcpy(odate,"Apr ");
  else   if (!strcmp(month,"05"))
    strcpy(odate,"May ");
  else   if (!strcmp(month,"06"))
    strcpy(odate,"Jun ");
  else   if (!strcmp(month,"07"))
    strcpy(odate,"Jul ");
  else   if (!strcmp(month,"08"))
    strcpy(odate,"Aug ");
  else   if (!strcmp(month,"09"))
    strcpy(odate,"Sep ");
  else   if (!strcmp(month,"10"))
    strcpy(odate,"Oct ");
  else   if (!strcmp(month,"11"))
    strcpy(odate,"Nov ");
  else   if (!strcmp(month,"12"))
    strcpy(odate,"Dec ");


  strncat(odate,&date[8],2);
  if (odate[4] == '0') odate[4] = ' ';

  strcat(odate," ");
  strncat(odate,&date[0],4);

}
#endif // VLAD_DATE_OBSOLETE


void Db::ConvertDate(string& dbDate, const string& cifDate)
{
    //
    // Reformat CIF date = yyyy-mm-dd[:00:00]  to  Sybase date yyyy-Mon-dd hh:mm
    // 2000-01-27  to 2000-Jan-27 00:00  odate[18]
    //

    dbDate.clear();

    // CIF date must be in the format YYYY[-MM[-DD]]
    if ((cifDate.size() != 4) && (cifDate.size() != 7) &&
      (cifDate.size() != 10))
    {
#ifdef VLAD_LOG_SEPARATION
        if (_verbose)
            _log << "ConvertDate() date format error: " << cifDate << endl;
#endif

        return;
    }

    if (cifDate.size() == 10)
    {
        // The input data is YYYY-MM-DD
        if ((cifDate[4] != '-') && (cifDate[7] != '-'))
        {
#ifdef VLAD_LOG_SEPARATION
            if (_verbose)
                _log << "ConvertDate() date format error: " << cifDate << endl;
#endif
        }
        else
        {
            dbDate = cifDate;
        }
    }
    else if (cifDate.size() == 7)
    {
        // The input data is YYYY-MM
        if (cifDate[4] != '-')
        {
#ifdef VLAD_LOG_SEPARATION
            if (_verbose)
                _log << "ConvertDate() date format error: " << cifDate << endl;
#endif
        }
        else
        {
            dbDate = cifDate + "-01";
        }
    }
    else
    {
        // The input data is YYYY
        dbDate = cifDate + "-01" + "-01";
    }

}


void Db::ConvertTimestamp(string& dbTimestamp, const string& cifTimestamp)
{

    dbTimestamp.clear();

    if (CifString::IsEmptyValue(cifTimestamp))
    {
        return;
    }

    ConvertDate(dbTimestamp, cifTimestamp);

    string::size_type firstColonPos = cifTimestamp.find(':', 0);

    if (firstColonPos == string::npos)
        dbTimestamp += " 00:00";
    else
        dbTimestamp += cifTimestamp.substr(firstColonPos + 1);

}


#ifdef VLAD_DATE_OBSOLETE
void DbSybase::ConvertDate(string& dbDate, const string& cifDate)
{
    //
    // Reformat CIF date = yyyy-mm-dd[:00:00] to Sybase date yyyy-Mon-dd hh:mm
    // 2000-01-27  to 2000-Jan-27 00:00  odate[18]
    //
    // int i, iDay, maxDay, year;
    //char month[5], day[3];
    int maxDay;

    dbDate.clear();

    string baseDate;
    Db::ConvertDate(baseDate, cifDate);

    // Put YYYY and -
    dbDate = baseDate.substr(0, 5);

    //string substr( size_type index, size_type length = npos );

    // Convert MM to textual month and add it
    string::size_type firstDashPos = baseDate.find('-', 0);
    string::size_type secondDashPos = baseDate.find('-', firstDashPos + 1);
  
    string month = baseDate.substr(firstDashPos + 1, secondDashPos -
      (firstDashPos + 1));

    if (month == "01")
    {
        dbDate += "Jan";
        maxDay = 31;
    }
    else if (month == "02")
    {
        dbDate += "Feb";
        int year = atoi((baseDate.substr(0, 4)).c_str());
        if (year % 4 == 0 && year % 100 != 0 || year % 400 == 0)
             maxDay = 29;
        else
             maxDay = 28;
    }
    else if (month == "03")
    {
      dbDate += "Mar";
      maxDay = 31;
    }
    else if (month == "04")
    {
      dbDate += "Apr";
      maxDay = 30;
    }
    else if (month == "05")
    {
        dbDate += "May";
        maxDay = 31;
    }
    else if (month == "06")
    {
        dbDate += "Jun";
        maxDay = 30;
    }
    else if (month == "07")
    {
        dbDate += "Jul";
        maxDay = 31;
    }
    else if (month == "08")
    {
        dbDate += "Aug";
        maxDay = 31;
    }
    else if (month == "09")
    {
        dbDate += "Sep";
        maxDay = 30;
    }
    else if (month == "10")
    {
        dbDate += "Oct";
      maxDay = 31;
    }
    else if (month == "11")
    {
        dbDate += "Nov";
        maxDay = 30;
    }
    else if (month == "12")
    {
        dbDate += "Dec";
        maxDay = 31;
    }
    else
    {
#ifdef VLAD_LOG_SEPARATION
        if (_verbose)
            _log << "_dformat3() month format error: " << cifDate << endl;
#endif

        dbDate.clear();

        return;
    }

    // Put day

    dbDate += baseDate.substr(secondDashPos, 3);

}


void DbLoader::_dformat_3(const char *date, char *odate, int shortFlag) {
  //
  // reformat  CIF date = yyyy-mm-dd[:00:00]  to  Sybase date yyyy-Mon-dd hh:mm
  //                      2000-01-27  to 2000-Jan-27 00:00  odate[18]
  //
  int i, iDay, maxDay, year;
  char month[5], day[3];

  if (odate == NULL) return;
  strcpy(odate,"");

  if (date == NULL ||  date[4] != '-' || date[7] != '-' ) {
    if (_verbose) _log << "_dformat3() date format error: " << date << endl;
    return;
  }


  memset(month, 0,5);
  strncpy(month,&date[5],2);


  strcpy (odate,"");
  strncat(odate,&date[0],4);

  for (i=0; i < 4; i++) {
    if (!isdigit(odate[i])) {
      if (_verbose) _log << "_dformat3() year format error: " << date << endl;
      strcpy(odate,"");
      return;
    }
  }

  year = atoi(odate);
  strcat(odate,"-");

  if (!strcmp(month,"01")) {
    strcat(odate,"Jan");
    maxDay = 31;
  }  else   if (!strcmp(month,"02")) {
    strcat(odate,"Feb");
    if (year % 4 == 0 && year % 100 != 0 || year % 400 == 0)
         maxDay = 29;
    else maxDay = 28;
  } else   if (!strcmp(month,"03")) {
    strcat(odate,"Mar");
    maxDay = 31;
  } else   if (!strcmp(month,"04")) {
    strcat(odate,"Apr");
    maxDay = 30;
  } else   if (!strcmp(month,"05")) {
    strcat(odate,"May");
    maxDay = 31;
  } else   if (!strcmp(month,"06")) {
    strcat(odate,"Jun");
    maxDay = 30;
  } else   if (!strcmp(month,"07")) {
    strcat(odate,"Jul");
    maxDay = 31;
  } else   if (!strcmp(month,"08")) {
    strcat(odate,"Aug");
    maxDay = 31;
  } else   if (!strcmp(month,"09")) {
    strcat(odate,"Sep");
    maxDay = 30;
  } else   if (!strcmp(month,"10")) {
    strcat(odate,"Oct");
    maxDay = 31;
  } else   if (!strcmp(month,"11")) {
    strcat(odate,"Nov");
    maxDay = 30;
  } else   if (!strcmp(month,"12")) {
    strcat(odate,"Dec");
    maxDay = 31;
  } else {  //
    if (_verbose) _log << "_dformat3() month format error: " << date << endl;
    strcpy(odate,"");
    return;
  }




  strcat(odate,"-");
  strncat(odate,&date[8],2);

  if (!isdigit(odate[9]) || !isdigit(odate[10])) {
    if (_verbose) _log << "_dformat3() day format error: " << date << endl;
    strcpy(odate,"");
    return;
  }
  
  day[0] = odate[9];
  day[1] = odate[10];
  day[2] = '\0';
  iDay = atoi(day);
  if ((iDay <= 0) || (iDay > maxDay)) {
    if (_verbose) _log << "_dformat3() day format error: " << date << endl;
    strcpy(odate,"");
    return;
  }

  if (!shortFlag) strcat(odate," 00:00");
}


void DbLoader::_dformat_4(const char *date, char *odate) {
  //
  // Mon dd yyyy to CIF date yyyy-mm-dd  odate[11]
  //
  if (odate == NULL) return;
  strcpy(odate,"");

  if (date == NULL ||  date[3] != ' ' || date[6] != ' ' ) return;

  char month[5];
  memset(month, 0 ,5);
  strncpy(month,date,3);


  strcpy (odate,"");
  strncat(odate,&date[7],4);
  strcat(odate,"-");

  if (!strcmp(month,"Jan"))
    strcat(odate,"01");
  else   if (!strcmp(month,"Feb"))
    strcat(odate,"02");
  else   if (!strcmp(month,"Mar"))
    strcat(odate,"03");
  else   if (!strcmp(month,"Apr"))
    strcat(odate,"04");
  else   if (!strcmp(month,"May"))
    strcat(odate,"05");
  else   if (!strcmp(month,"Jun"))
    strcat(odate,"06");
  else   if (!strcmp(month,"Jul"))
    strcat(odate,"07");
  else   if (!strcmp(month,"Aug"))
    strcat(odate,"08");
  else   if (!strcmp(month,"Sep"))
    strcat(odate,"09");
  else   if (!strcmp(month,"Oct"))
    strcat(odate,"10");
  else   if (!strcmp(month,"Nov"))
    strcat(odate,"11");
  else   if (!strcmp(month,"Dec"))
    strcat(odate,"12");

  strcat(odate,"-");
  strncat(odate,&date[4],2);
  if (odate[8] ==  ' ') odate[8] = '0';

}




void DbLoader::_dformat_5(const char *date, char *odate) {
  //
  // reformat  CIF date = yyyy[-mm-dd:00:00]  to  Oracle/DB2 date yyyy-mm-dd
  //                      2000-01  to 2000-01-01   odate[10]
  //
  int i;

  if (odate == NULL) return;
  strcpy(odate,"");

  if (date == NULL || (strlen(date) < 4) ) {
    if (_verbose) _log << "_dformat_5() date format error" << endl;
    return;
  }

  // Year 
  strncat(odate,&date[0],4);
  for (i=0; i < 4; i++) {
    if (!isdigit(odate[i])) {
      if (_verbose) _log << "_dformat_5() year format error: " << date << endl;
      strcpy(odate,"");
      return;
    }
  }

  //month and day

  if (strlen(date) == 4) {
    strcat(odate,"-01-01");
    return;
  } else if ((strlen(date) == 7) && isdigit(date[5]) && isdigit(date[6])) {
    strcat(odate,"-");
    strncat(odate,&date[5],2);
    strcat(odate,"-01");
    return;
  } else if ((strlen(date) >= 10) && isdigit(date[8]) && isdigit(date[9])) {
    strcat(odate,"-");
    strncat(odate,&date[5],2);
    strcat(odate,"-");
    strncat(odate,&date[8],2);
    return;
  }
}
#endif // VLAD_DATE_OBSOLETE



void DbLoader::_ReorderName(string& res, char *aString, int mode)
{  
  //  mode = 1. Reformat name:  Lastname, I.J. -> I.J. Lastname
  //  mode = 2. Reformat name:  I.J. Lastname  -> Lastname, I.J. 
  
  res.clear();

  if (aString == NULL || aString[0] == '\0') return;
  
  int i, ilen = 0, c='\0';
  ilen = strlen(aString);
  if (ilen < 1) return;


  char *first = new char[ilen+1]; 
  char *last  = new char[ilen+1]; 
  
  if (mode == 1) {
    for (i = strlen(aString) - 1; i >= 0; i--)
      if (aString[i] == ',') break;
    
    if (i > 0) {
      strcpy(first, &aString[i+1]);
      aString[i] = '\0';
      strcpy(last, aString);
      string firstClean = first;
      string lastClean = last;
      CleanString(firstClean);
      CleanString(lastClean);
      if (firstClean[firstClean.size()-1] == '.')
	sprintf(aString, "%s%s%c", first, last,c);
      else 
	sprintf(aString, "%s %s%c", first, last, c);
    }
    delete first;
    delete last;
    res = aString;
    return;
    
  } else if (mode == 2) {
    for (i = strlen(aString) - 1; i >= 0; i--) {
      if (aString[i] == '.') break;
    }

    if (i > 0) {
      strcpy(last, &aString[i+1]);
      aString[i+1] = '\0';
      strcpy(first, aString);
      sprintf(aString, "%s, %s%c", last, first,c);
      res = aString;
      CleanString(res);
      delete first;
      delete last;
      return;
    }
    
    for (i = strlen(aString) - 1; i >= 0; i--)
      if (aString[i] == ' ') break;
    
    if (i > 0) {
      strcpy(last, &aString[i+1]);
      aString[i+1] = '\0';
      strcpy(first, aString);
      sprintf(aString, "%s, %s%c", last, first, c);
      res = aString;
      CleanString(res);
    }
    delete first;
    delete last;
    return;
  }
}  


void DbLoader::CleanString(string& aString)
{
/*
 *  CleanString() strip leading and trailing white space from a string.
 *  Also compress multiple internal white space characters.
 *
 */

  int i, stringlen;
  string res;

  if (aString.empty())
      return;

  stringlen = aString.size();

  for (i=0; i< stringlen; i++)
  { 
      if (aString[i] == ' ' || aString[i] == '\t' ||  aString[i] == '\n')
      {
          //strcpy(aString, &aString[i+1]);	
          //stringlen--;
          //i--;
      }
      else
          break;
  }

  int savedIndex = i;

  for (/*i=0*/; i< stringlen; i++)
  {
      if (i != savedIndex && 
	(aString[i-1] == ' ' || aString[i-1] == '\t' || aString[i-1] == '\n')  && 
	(aString[i]   == ' ' || aString[i]   == '\t' || aString[i]   == '\n' ))       {
          //strcpy(&aString[i-1], &aString[i]);
          //i--;
          //stringlen--;
      }
      else
      {
          if (aString[i] == '\t' || aString[i] == '\n' )
              res.push_back(' ');
          else
              res.push_back(aString[i]);
      }
  }

#ifdef VLAD_DELETED
  stringlen = strlen(aString);
  for (i=0; i< stringlen; i++)
  {
      if (aString[i] == '\t' || aString[i] == '\n' )
      {
          aString[i] = ' ';
      }
  }
#endif

  for (unsigned int i=0; i < res.size(); ++i)
  {
    if (res[res.size() - 1 - i] == ' ' || res[res.size() - 1 - i] == '\t' ||  res[res.size() - 1 - i] == '\n')
        res.erase(res.end() - 1);
    else
        break;
  }

  aString = res;

}


void DbLoader::_ToUpperString(string& aString)
{
/*
 *  _ToUpperString() set string to upper case.
 *
 */

  int i;
  if (_verbose) _log << "_ToUpperString: " << aString << endl;
  if (aString.empty()) return;
  for (i=0; i< (int) aString.size(); i++) { 
    aString[i] = toupper(aString[i]);
  }
  if (_verbose) _log << "_ToUpperString: " << aString << endl;
}

void DbLoader::_StripString(string& aString, int mode)
{
/*
 *  _StripString() remove white space characters.  
 *        mode = 1  - all white space
 *        mode = 2  - remove only \n and \r
 */

  int i, j, iLen;
  char *ostring=NULL;
  if (_verbose) _log << "_StripString: input " << aString << endl;

  if (aString.empty()) return;

  iLen = aString.size();
  ostring = new char[iLen + 1];
  memset(ostring,'\0',iLen+1);

  j=0;
  if (mode == 1) {
    for (i=0; i< iLen; i++) { 
      if ( isspace(aString[i]) ) continue;
      ostring[j] = aString[i]; j++;
    }
  } else if (mode == 2) {
    for (i=0; i< iLen; i++) { 
      if ( aString[i] == '\n' || aString[i] == '\r'  ) continue;
      ostring[j] = aString[i]; j++;
    }

  }
  aString.clear();
  aString = ostring;
  if (_verbose) _log << "_StripString: output " << aString << endl;
  if (ostring) delete[] ostring;

}


static void escapeString(string& outStr, const string& inStr)
{

  int i, iLen;

  outStr.clear();

  if (inStr.empty())
      return;
  
  iLen = inStr.size();

  for (i=0; i < iLen; i++)
  {
    if (inStr[i] == '\r')
    {
      outStr.push_back('\\');
      outStr.push_back('r');
    }
    else if (inStr[i] == '\n')
    {
      outStr.push_back('\\');
      outStr.push_back('n');
    }
    else if (inStr[i] == '\t')
    {
      outStr.push_back('\\');
      outStr.push_back('t');
    }
    else 
      outStr.push_back(inStr[i]);
  }
  
}


static void _set_attribute_value_where(ISTable* t, const string& tableName,
  const string& attribName, const string& attributeChange,
  const string& valueChange)

{

    if ((t == NULL) || attribName.empty() || tableName.empty() ||
      attributeChange.empty() || valueChange.empty())
        return;

    if (!(t->IsColumnPresent(attributeChange)))
    {
        cout << "bad attribute in name" << endl;
        return;
    }

    vector<unsigned int> ir;
    vector<string> qcol, qval;    

    qcol.push_back("table_name");
    qcol.push_back("attribute_name");

    qval.push_back(tableName);
    qval.push_back(attribName);

    t->Search(ir, qval, qcol);

    if (!ir.empty())
    {
        t->UpdateCell(ir[0], attributeChange, valueChange);
    }

}

unsigned int DbLoader::GetTableColumnIndex(const ISTable& isTable,
  const string& colName)
{

    if (colName.empty())
    {
        // Empty column name.
        throw EmptyValueException("Empty column name",
          "DbLoader::GetTableColumnIndex");
    }

    const vector<string>& colNames = isTable.GetColumnNames();

    unsigned int colIndex = 0;

    for (colIndex = 0; colIndex < colNames.size(); ++colIndex)
    {
        if (String::IsEqual(colNames[colIndex], colName,
          isTable.GetColCaseSense()))
        {
            break;
        }
    }

    if (colIndex == colNames.size())
    {
        // Column not found.
        throw NotFoundException("Column not found",
          "DbLoader::GetTableColumnIndex");
    }
    else
    {
        return(colIndex);
    }

}


static void name_conversion_first_last(char* aString)
{
    int i, len;
    char first[500], last[500];

    if (aString == NULL || aString[0] == '\0')
        return;

    len = strlen(aString);
    for (i = len - 1; i >= 0; i--)
        if (aString[i] == ',')
            break;

    if (i > 0)
    {
        strcpy(first, &aString[i+1]);
        aString[i] = '\0';
        strcpy(last, aString);
        ndb_clean_string(first);
        ndb_clean_string(last);
        if (first[strlen(first)-1] == '.')
            sprintf(aString, "%s%s", first, last);
        else
            sprintf(aString, "%s %s", first, last);
    }
}


#ifdef DB_HASH_ID
long long DbLoader::pdbIdHash(const string& id) 
{

    long long ival, base, base1;

    int i, j;
    int id1[5];

    const char* list = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"; 

    ival = (long long) 0;
    base = (long long) strlen(list);

    if (id.empty() || (id.size() < 4))
        return ival;

    if (id.size() == 4)
    {
        base1 = (long long) 1;
        for (i = 3, j = 0; i >= 0; i--)
        {
            if (i != 3)
                base1 = base1*base;
      
            for (j = 0; j < base; j++)
            {
	        if (id[i] == list[j])
                {
                    ival = ival + base1 * j;
                    break;
                }
            }
        }
        ival = ival * 4000000;
    }
    else if (id.size() == 7)
    {
        i=0;
        for (j = 3; j < 7; j++)
        {
            id1[i] = id[j];
            i++;
        }
        id1[i] = '\0';

        base1 = (long long) 1;
        for (i = 3, j = 0; i >= 0; i--)
        {
            if (i != 3)
                base1 = base1*base;
      
            for (j = 0; j < base; j++)
            {
	        if (id1[i] == list[j])
                {
                    ival = ival + base1 * j;
                    break;
                }
            }
        }
        ival = ival * 4000000;
    }
    else if (id.size() == 10)
    {
        base = 10;
        base1 = (long long) 1;
        for (i = 9, j = 0; i >= 4; i--)
        {
            if (i != 9)
                base1 = base1*base;
 
            for (j = 0; j < base; j++)
            {
	        if (id[i] == list[j])
                {
                    ival = ival + base1 * j;
                    break;
                }
            }
        }
        ival = ival + 2000000;
        ival = ival * 4000000;
    }
    else if (id.size() == 6)
    {
        // NDB ids without pdb ids.
        base = (long long) strlen(list);
        base1 = (long long) 1;
        for (i = 5, j = 0; i >= 0; i--)
        {
            if (i != 5)
                base1 = base1*base;
      
            for (j = 0; j < base; j++)
            {
                if (id[i] == list[j])
                {
                    ival = ival + base1 * j;
                    break;
                }
            }
        }
        ival = ival + 2000000;
        ival = ival * 2000000;
    }

    return (ival);

}
#endif
